#include "stdafx.h"
#include "body/main.h"

//////////////////////////////////////////////////////////////////////
////  BK::W32FPSGear
//

namespace {

inline DWORD get_diff ( DWORD prev, DWORD curr )
{
    return (curr < prev) ? (UINT_MAX - prev) + 1 + curr : curr - prev;
}

} /* namespace */

BK::W32FPSGear::W32FPSGear ( void )
        : num_last_time (0),
          num_last_count (0),
          num_last_fps (0.0)
{
}

BK::W32FPSGear::~W32FPSGear ()
{
}

void BK::W32FPSGear::execute ( void )
{
    GraphicDevice * const gdev = GraphicDevice::get_instance();
    const unsigned int prevcount = num_last_count;
    const unsigned int currcount = (gdev == NULL) ? num_last_count : gdev->get_screen_update_count();
    const DWORD prevtime = num_last_time;
    const DWORD currtime = timeGetTime();
    
    if( prevtime == 0 ) {
        num_last_time = currtime;
        num_last_count = currcount;
        return;
    }
    
    const DWORD diff = get_diff( prevtime, currtime );
    
    if( diff >= get_interval() ) {
        num_last_fps = static_cast< double >(currcount - prevcount) / diff;
        num_last_time = currtime;
        num_last_count = currcount;
    }
}

double BK::W32FPSGear::get_fps ( void ) const
{
    return num_last_fps;
}

//////////////////////////////////////////////////////////////////////
