#include "stdafx.h"
#include "body/main.h"

//////////////////////////////////////////////////////////////////////
////  t@C[JȒ萔
//
namespace {

using namespace BK;

const W32FrameTimer::schedule_t SCHEDULE_30 ( 30, 33, 33, 34, 0 );
const W32FrameTimer::schedule_t SCHEDULE_60 ( 60, 16, 17, 17, 0 );

const W32FrameTimer::schedule_t * const PRESET_TIMING [] = 
{
    &SCHEDULE_30,
    &SCHEDULE_60,
    NULL
};

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  t@C[Jȃ\bh
//
namespace {

using namespace BK;

/// w肳ꂽ FPS ̃^C}[f[^擾
const W32FrameTimer::schedule_t * get_builtin_schedule ( unsigned int fps )
{
    const W32FrameTimer::schedule_t * const * ptr = PRESET_TIMING;
    
    while( *ptr != NULL ) {
        const W32FrameTimer::schedule_t * const schedule = *(ptr++);
        
        if( schedule->get_fps() == fps ) {
            return schedule;
        }
    }
    
    return NULL;
}

/// VXěݎ擾
inline FrameTimer::time_t get_system_time ( MMTIME * ptr )
{
    timeGetSystemTime( ptr, sizeof(MMTIME) );
    
    return static_cast< time_t >( ptr->u.ms );
}

/// VXěݎ擾
inline FrameTimer::time_t get_system_time ( void )
{
    MMTIME buf;
    
    buf.wType = TIME_MS;
    return get_system_time( &buf );
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::W32FrameTimer
//

BK::W32FrameTimer::schedule_t::schedule_t ( unsigned int fps, ... )
        : num_fps (fps),
          num_length (0),
          num_distance (0)
{
    va_list arg;
    unsigned int n;
    
    va_start( arg, fps );
    while( (n = va_arg(arg, unsigned int)) != 0 ) {
        if( this->num_length >= sizeof(this->ary_table)/sizeof(this->ary_table[0]) ) {
            throw new logic_exception("W32FrameTimer::schedule_t ̍\zɓnꂽt[܂. (%d FPS)", this->num_fps);
        }
        this->ary_table[ this->num_length ] = n;
        this->num_length += 1;
        this->num_distance += n;
    }
    va_end( arg );
}

BK::W32FrameTimer::W32FrameTimer ( unsigned int fps, BK::FrameTimer::time_t curr )
        : FrameTimer (fps, curr),
          ptr_schedule (get_builtin_schedule(fps)),
          num_interval_index (0)
{
    this->obj_mmtime.wType = TIME_MS;
    
    if( this->ptr_schedule == NULL ) {
        throw new argument_exception("FPS %u ̃t[^C}[͔T|[gł.", fps);
    }
}

BK::W32FrameTimer::~W32FrameTimer ()
{
}

BK::FrameTimer::time_t BK::W32FrameTimer::get_current_time ( void )
{
    return get_system_time( &this->obj_mmtime );
}

//////////////////////////////////////////////////////////////////////
////  BK::W32CPUFrameTimer
//

BK::W32CPUFrameTimer::W32CPUFrameTimer ( unsigned int fps )
        : W32FrameTimer (fps, get_system_time()),
          num_last_time (0)
{
    if( log_is_info() ) {
        log_info("%u FPS CPUt[^C}[\z.", fps);
    }
}

BK::W32CPUFrameTimer::~W32CPUFrameTimer ()
{
    if( log_is_info() ) {
        log_info("CPUt[^C}[j.");
    }
}

BK::FrameTimer::time_t BK::W32CPUFrameTimer::on_wait_frame ( void )
{
    const schedule_t * const schedule = this->get_schedule();
    time_t lasttime = this->num_last_time;
    time_t currtime;
    time_t nexttime;
    
    if( lasttime == 0 ) {
        lasttime = this->get_current_time();
    }
    currtime = this->get_current_time();
    
    // t[̎߂
    nexttime = lasttime;
    {
        // ȏ㗣Ăꍇ͐p̌vZs
        const unsigned int distance = schedule->get_distance();
        
        if( lasttime - currtime >= distance ) {
            nexttime += ((lasttime - currtime) / distance) * distance;
        }
        while( nexttime < currtime ) {
            nexttime += schedule->get_interval( this->advance_interval_index() );
        }
    }
    
    // t[܂ő҂
    {
        // ҂Ԃ\ɑ傫 (Windows ł 20ms) ꍇ̓X[v
        currtime = this->get_current_time();
        if( nexttime - currtime >= 20 + 5 ) {
            Sleep( nexttime - currtime - 5 );
            currtime = this->get_current_time();
        }
        
        // ݎI[o[t[Ƃ͍̎lȂc
        while( currtime < nexttime ) {
            currtime = this->get_current_time();
        }
    }
    
    this->num_last_time = nexttime;
    return nexttime;
}

//////////////////////////////////////////////////////////////////////
