#include "stdafx.h"
#include "body/main.h"

//////////////////////////////////////////////////////////////////////
////  BK::Dx9SurfaceCanvas
//
namespace BK {

/// Direct3D T[tFXpLoX
/**
 * Direct3D T[tFXp̃LoXIuWFNgłB
 * bNT[tFX̃̈gp܂B
 *
 * LoXɃT[tFXĂ͂Ȃ܂B
 * ܂ALoX݂ԁAT[tFX݂ĂȂ΂Ȃ܂B
 */
class Dx9SurfaceCanvas : public BK::Canvas
{
public: /* Canvas C^[tFCX */
    
    /// LoX̃N[Y
    virtual void close_canvas ( void );
    
private:
    
    /// I[i[ƂȂT[tFX
    Dx9Surface & obj_owner;
    
public:
    
    /// RXgN^
    /**
     * T[tFX̃bN̈LoX\z܂B
     *
     * @param owner [in] T[tFXIuWFNg (bNς݂łȂ΂ȂȂ)
     * @param lock [in] bN
     * @param area [in] bNΏۂ͈̔͏
     */
    Dx9SurfaceCanvas ( Dx9Surface & owner, const D3DLOCKED_RECT & lock, const Rect2 & area );
    
    /// fXgN^
    virtual ~Dx9SurfaceCanvas ();
    
};

} /* namespace BK */

BK::Dx9SurfaceCanvas::Dx9SurfaceCanvas ( BK::Dx9Surface & owner, const D3DLOCKED_RECT & lock, const Rect2 & area )
        : Canvas (&owner, area, owner.get_pixel_info(), lock.pBits, lock.Pitch),
          obj_owner (owner)
{
}

BK::Dx9SurfaceCanvas::~Dx9SurfaceCanvas ()
{
    if( !is_canvas_closed() ) {
        close_canvas();
    }
}

void BK::Dx9SurfaceCanvas::close_canvas ( void )
{
    obj_owner.unlock_surface();
    Canvas::close_canvas();
}

//////////////////////////////////////////////////////////////////////
////  t@C[Jȃ\bh
//
namespace BK {

/// Dx9SurfaceCanvas ̐
/**
 * Canvas RXgN^ɕKvȃp[^߁Ã\bhŎOɏ܂B
 */
Dx9SurfaceCanvas * create_Dx9SurfaceCanvas ( Dx9Surface & owner, const Rect2 * areaptr )
{
    Rect2  area;
    RECT * pwr;
    RECT   wr;
    
    if( areaptr == NULL ) {
        area = Rect2( owner.get_size() );
        pwr = NULL;
    }
    else {
        area = *areaptr;
        wr = toRECT(area);
        pwr = &wr;
    }
    
    D3DLOCKED_RECT lock = owner.lock_surface(pwr);
    try {
        return new Dx9SurfaceCanvas( owner, lock, area );
    } catch ( exception_base * ) {
        owner.unlock_surface();
        throw;
    }
}

} /* namespace BK */
//////////////////////////////////////////////////////////////////////
////  t@C[Jȃ\bh
//
namespace {

using namespace BK;

IDirect3DSurface9 * create_direct3d_surface ( const Size2 & size, D3DFORMAT format, bool vram )
{
    Dx9GraphicDevice * const gdev = Dx9GraphicDevice::get_instance();
    IDirect3DSurface9 * surface;
    
    if( gdev == NULL ) {
        throw new logic_exception("sN`̐ɂ̓OtBbNfoCXKvł.");
    }
    
    HRESULT result;
    D3DPOOL pool;
    
    if( vram ) {
        pool = D3DPOOL_MANAGED;
    } else {
        pool = D3DPOOL_SYSTEMMEM;
    }
    
    result = gdev->CreateOffscreenPlainSurface( size, format, pool, &surface );
    if( result != D3D_OK ) {
        throw new directx_exception(result, "sN`̐Ɏs.");
    }
    
    return surface;
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::Dx9Surface
//

void BK::Dx9Surface::init ( void )
{
    if( ptr_surface == NULL ) {
        throw new logic_exception("Dx9Surface IuWFNg NULL ŏł܂.");
    }
    
    D3DSURFACE_DESC desc;
    
    DirectX::get_d3d_surface_desc( ptr_surface, &desc );
    
    num_pool = desc.Pool;
    num_format = desc.Format;
}

BK::Dx9Surface::Dx9Surface ( IDirect3DSurface9 * surface )
        : ptr_surface (surface),
          num_pool (D3DPOOL_MANAGED),
          num_format (D3DFMT_UNKNOWN)
{
    init();
}

BK::Dx9Surface::Dx9Surface ( BK::Size2 const & size, D3DFORMAT format, bool vram )
        : ptr_surface (create_direct3d_surface(size, format, vram)),
          num_pool (D3DPOOL_MANAGED),
          num_format (D3DFMT_UNKNOWN)
{
    try {
        init();
        if( log_is_trace() ) {
            switch( num_pool ) {
            case D3DPOOL_DEFAULT:
                log_trace("Direct3D T[tFX <%p>  VRAM ɍ쐬܂.", ptr_surface);
                break;
                
            case D3DPOOL_SYSTEMMEM:
                log_trace("Direct3D T[tFX <%p> VXeɍ쐬܂.", ptr_surface);
                break;
            }
            log_write_trace( DirectX::inspect_d3d_surface(ptr_surface) );
        }
    } catch ( exception_base * ) {
        ptr_surface->Release();
        throw;
    }
}

BK::Dx9Surface::~Dx9Surface ()
{
    DirectX::Release(ptr_surface);
    if( log_is_trace() ) {
        log_trace("Direct3D T[tFX <%p> ܂.", ptr_surface);
    }
}

BK::Size2 BK::Dx9Surface::get_size ( void )
{
    D3DSURFACE_DESC desc;
    
    DirectX::get_d3d_surface_desc( ptr_surface, &desc );
    
    return Size2( desc.Width, desc.Height );
}

std::string BK::Dx9Surface::inspect ( void ) const
{
    return DirectX::inspect_d3d_surface(ptr_surface);
}

BK::Bitmap::pool_t BK::Dx9Surface::get_pool_info ( void ) const
{
    return DirectX::D3DPOOL_to_pool( num_pool );
}

BK::Canvas::pixel_t BK::Dx9Surface::get_pixel_info ( void ) const
{
    return DirectX::D3DFORMAT_to_pixel( num_format );
}

BK::Canvas * BK::Dx9Surface::create_canvas ( BK::Rect2 const * area )
{
    return create_Dx9SurfaceCanvas(*this, area);
}

void BK::Dx9Surface::blit ( BK::Point2 const & dest, BK::Bitmap & src, BK::Rect2 const & area )
{
    IDirect3DSurface9 * const destsurface = get_d3d_surface();
    IDirect3DSurface9 * srcsurface = NULL;
    type_info const & srctype = typeid(src);
    HRESULT result;
    
    if( srctype == typeid(BK::Dx9Surface) ) {
        srcsurface = (dynamic_cast<BK::Dx9Surface &>(src)).get_d3d_surface();
    }
    else if( srctype == typeid(BK::Dx9Texture) ) {
        srcsurface = (dynamic_cast<BK::Dx9Texture &>(src)).get_d3d_surface();
    }
    else if( srctype == typeid(BK::Dx9SurfaceData) ) {
        srcsurface = (dynamic_cast<BK::Dx9SurfaceData &>(src)).get_surface()->get_d3d_surface();
    }
    
    if( srcsurface != NULL ) {
        const Size2 srcsize (src.get_size());
        const Rect2 srcrect (srcsize);
        
        if( !srcrect.is_include(area) ) {
            throw new argument_exception(
                "rbg}bv]œ]rbg}bv (%s) ̗̈O (%s) w肳܂.",
                srcrect.to_string().c_str(), area.to_string().c_str());
        }
        
        Rect2 destarea (dest, srcsize);
        destarea.clipping( get_size() );
        
        RECT destwr = toRECT(destarea);
        RECT srcwr = toRECT(area);
        
        result = D3DXLoadSurfaceFromSurface(
            destsurface,
            NULL,
            &destwr,
            srcsurface,
            NULL,
            &srcwr,
            D3DX_FILTER_NONE,
            0 );
        if( result != D3D_OK ) {
            throw new directx_exception(result, "Direct3D T[tFX̓]Ɏs.");
        }
    }
    else {
        throw new unsupported_exception(
            "Direct3D IuWFNgȊO (%s)  Direct3D T[tFXւ̓]̓T|[gĂ܂.",
            srctype.name());
    }
}

void BK::Dx9Surface::write_to_file ( const char * filename )
{
    Dx9GraphicDevice * const gdev = Dx9GraphicDevice::get_instance();
    
    if( gdev == NULL ) {
        throw new logic_exception("VXe DirectGraphics IuWFNgȂ.");
    }
    
    HRESULT result;
    
    if( filename == NULL ) {
        throw new argument_exception("Direct3D T[tFX̏o͐t@C NULL w肳܂.");
    }
    
    if( get_d3d_format() != D3DFMT_A8R8G8B8 ) {
        // ԃT[tFX (A8R8G8B8 ɃRs[Ă)
        Dx9Surface tempsurface (get_size(), D3DFMT_A8R8G8B8, false);
        
        tempsurface.blit( Point2(0, 0), *this, Rect2(get_size()) );
        
        result = D3DXSaveSurfaceToFile( filename, D3DXIFF_BMP, tempsurface.get_interface(), NULL, NULL );
        if( result != D3D_OK ) {
            throw new directx_exception(result, "t@C '%s' ւ̒ԃT[tFX`݂Ɏs.", filename);
        }
        log_trace("%s ϊԃT[tFXt@C '%s' ɏ݂܂.",
                  DirectX::D3DFORMAT_to_string(get_d3d_format()), filename);
    }
    else {
        result = D3DXSaveSurfaceToFile( filename, D3DXIFF_BMP, get_interface(), NULL, NULL );
        if( result != D3D_OK ) {
            throw new directx_exception(result, "t@C '%s' ւ Direct3D T[tFX`݂Ɏs.", filename);
        }
        log_trace("Direct3D T[tFX <%p> t@C '%s' ɏ݂܂.", this, filename);
    }
}

D3DLOCKED_RECT BK::Dx9Surface::lock_surface ( RECT * r )
{
    D3DLOCKED_RECT lock;
    HRESULT result;
    
    result = get_interface()->LockRect( &lock, r, 0 );
    if( result != D3D_OK ) {
        if( r == NULL ) {
            throw new directx_exception(result, "Direct3D T[tFX̃bNɎs. (pRect=NULL)");
        } else {
            throw new directx_exception(result, "Direct3D T[tFX̃bNɎs. (pRect=(%d,%d,%d,%d))",
                                        r->left, r->top, r->right, r->bottom);
        }
    }
    #if 0
    if( log_is_trace() ) {
        log_trace("Direct3D T[tFX <%p> bN܂.", get_interface());
    }
    #endif
    return lock;
}

void BK::Dx9Surface::unlock_surface ( void )
{
    HRESULT result;
    
    result = get_interface()->UnlockRect();
    if( result != D3D_OK ) {
        throw new directx_exception(result, "Direct3D T[tFX̃bNɎs.");
    }
    #if 0
    if( log_is_trace() ) {
        log_trace("Direct3D T[tFX <%p> ̃bN܂.", get_interface());
    }
    #endif
}

//////////////////////////////////////////////////////////////////////
