#include "stdafx.h"
#include "body/main.h"

//////////////////////////////////////////////////////////////////////
////  BK::Dx9TextureCanvas
//
namespace BK {

class Dx9TextureCanvas : public BK::Canvas
{
public: /* Canvas C^[tFCX */
    
    /// LoX̃N[Y
    virtual void close_canvas ( void );
    
private:
    
    /// I[i[ƂȂeNX`
    Dx9Texture & obj_owner;
    
public:
    
    /// RXgN^
    /**
     * eNX`̃bN̈LoX\z܂B
     * KAx 0 ̃eNX`\[XbNȂ΂Ȃ܂B
     *
     * @param owner [in] eNX`IuWFNg (bNς݂łȂ΂ȂȂ)
     * @param lock [in] bN
     * @param area [in] bNΏۂ͈̔͏
     */
    Dx9TextureCanvas ( Dx9Texture & owner, const D3DLOCKED_RECT & lock, const Rect2 & area );
    
    /// fXgN^
    virtual ~Dx9TextureCanvas ();
    
};

Dx9TextureCanvas * create_Dx9TextureCanvas ( BK::Dx9Texture & owner, const BK::Rect2 * areaptr );

} /* namespace BK */

BK::Dx9TextureCanvas::Dx9TextureCanvas ( BK::Dx9Texture & owner, const D3DLOCKED_RECT & lock, const Rect2 & area )
        : Canvas (&owner, area, owner.get_pixel_info(), lock.pBits, lock.Pitch),
          obj_owner (owner)
{
}

BK::Dx9TextureCanvas::~Dx9TextureCanvas ()
{
    if( !is_canvas_closed() ) {
        close_canvas();
    }
}

void BK::Dx9TextureCanvas::close_canvas ( void )
{
    obj_owner.unlock_surface();
    Canvas::close_canvas();
}

//////////////////////////////////////////////////////////////////////
////  t@C[Jȃ\bh
//
namespace BK {

BK::Dx9TextureCanvas * BK::create_Dx9TextureCanvas ( BK::Dx9Texture & owner, const BK::Rect2 * areaptr )
{
    Rect2  area;
    RECT * pwr;
    RECT   wr;
    
    if( areaptr == NULL ) {
        area = Rect2( owner.get_size() );
        pwr = NULL;
    }
    else {
        area = *areaptr;
        wr = toRECT(area);
        pwr = &wr;
    }
    
    D3DLOCKED_RECT lock = owner.lock_surface(pwr);
    try {
        return new Dx9TextureCanvas( owner, lock, area );
    } catch ( exception_base * ) {
        owner.unlock_surface();
        throw;
    }
}

} /* namespace BK */
//////////////////////////////////////////////////////////////////////
////  t@C[Jȃ\bh
//
namespace {

using namespace BK;

IDirect3DTexture9 * create_direct3d_texture ( const Size2 & size, D3DFORMAT format, bool vram )
{
    Dx9GraphicDevice * const gdev = Dx9GraphicDevice::get_instance();
    IDirect3DTexture9 * texture;
    
    if( gdev == NULL ) {
        throw new logic_exception("C[W̐ɂ̓OtBbNfoCXKvł.");
    }
    
    HRESULT result;
    D3DPOOL pool;
    
    if( vram ) {
        pool = D3DPOOL_MANAGED;
    } else {
        pool = D3DPOOL_SYSTEMMEM;
    }
    
    result = gdev->CreateLv1Texture( size, format, pool, &texture );
    if( result != D3D_OK ) {
        throw new directx_exception(result, "C[W̐Ɏs.");
    }
    
    return texture;
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::Dx9Texture
//

void BK::Dx9Texture::init ( void )
{
    if( ptr_texture == NULL ) {
        throw new logic_exception("Dx9Texture IuWFNg NULL ŏł܂.");
    }
    
    IDirect3DSurface9 * surface;
    HRESULT result;
    
    result = ptr_texture->GetSurfaceLevel( 0, &surface );
    if( result != D3D_OK ) {
        throw new directx_exception(result, "eNX`T[tFX̎擾Ɏs.");
    }
    if( surface == NULL ) {
        throw new runtime_exception("eNX`T[tFX NULL Ԃ܂.");
    }
    ptr_surface = new Dx9Surface( surface );
    
    num_pool = ptr_surface->get_d3d_pool();
    num_format = ptr_surface->get_d3d_format();
}

BK::Dx9Texture::Dx9Texture ( IDirect3DTexture9 * texture )
        : ptr_texture (texture),
          ptr_surface (NULL),
          num_pool (D3DPOOL_MANAGED),
          num_format (D3DFMT_UNKNOWN)
{
    init();
}

BK::Dx9Texture::Dx9Texture ( BK::Size2 const & size, D3DFORMAT format, bool vram )
        : ptr_texture (create_direct3d_texture(size, format, vram)),
          ptr_surface (NULL),
          num_pool (D3DPOOL_MANAGED),
          num_format (D3DFMT_UNKNOWN)
{
    try {
        init();
        if( log_is_trace() ) {
            switch( num_pool ) {
            case D3DPOOL_MANAGED:
                log_trace("Direct3D eNX` <%p>  VRAM ɍ쐬܂.", ptr_texture);
                break;
                
            case D3DPOOL_SYSTEMMEM:
                log_trace("Direct3D eNX` <%p> VXeɍ쐬܂.", ptr_texture);
                break;
            }
            log_write_trace( DirectX::inspect_d3d_texture(ptr_texture) );
        }
    }
    catch ( exception_base * ) {
        ptr_texture->Release();
        delete ptr_surface;
        throw;
    }
}

BK::Dx9Texture::~Dx9Texture ()
{
    delete ptr_surface;
    ptr_texture->Release();
    if( log_is_trace() ) {
        log_trace("Direct3D eNX` <%p> ܂.", ptr_texture);
    }
}

BK::Size2 BK::Dx9Texture::get_size ( void )
{
    return ptr_surface->get_size();
}

std::string BK::Dx9Texture::inspect ( void ) const
{
    return DirectX::inspect_d3d_texture(ptr_texture);
}

BK::Bitmap::pool_t BK::Dx9Texture::get_pool_info ( void ) const
{
    return DirectX::D3DPOOL_to_pool( num_pool );
}

BK::Canvas::pixel_t BK::Dx9Texture::get_pixel_info ( void ) const
{
    return DirectX::D3DFORMAT_to_pixel( num_format );
}

BK::Canvas * BK::Dx9Texture::create_canvas ( BK::Rect2 const * area )
{
    return create_Dx9TextureCanvas(*this, area);
}

void BK::Dx9Texture::blit ( BK::Point2 const & dest, BK::Bitmap & src, BK::Rect2 const & area )
{
    ptr_surface->blit( dest, src, area );
}

void BK::Dx9Texture::write_to_file ( const char * filename )
{
    ptr_surface->write_to_file( filename );
}

D3DLOCKED_RECT BK::Dx9Texture::lock_surface ( RECT * r )
{
    D3DLOCKED_RECT lock;
    HRESULT result;
    
    result = get_interface()->LockRect( 0, &lock, r, 0 );
    if( result != D3D_OK ) {
        if( r == NULL ) {
            throw new directx_exception(result, "Direct3D eNX`̃bNɎs. (pRect=NULL)");
        } else {
            throw new directx_exception(result, "Direct3D eNX`̃bNɎs. (pRect=(%d,%d,%d,%d))",
                                        r->left, r->top, r->right, r->bottom);
        }
    }
    #if 0
    if( log_is_trace() ) {
        log_trace("Direct3D eNX` <%p> bN܂.", get_interface());
    }
    #endif
    return lock;
}

void BK::Dx9Texture::unlock_surface ( void )
{
    HRESULT result;
    
    result = get_interface()->UnlockRect( 0 );
    if( result != D3D_OK ) {
        throw new directx_exception(result, "Direct3D eNX`̃bNɎs.");
    }
    #if 0
    if( log_is_trace() ) {
        log_trace("Direct3D eNX` <%p> ̃bN܂.", get_interface());
    }
    #endif
}

//////////////////////////////////////////////////////////////////////
