#include "stdafx.h"
#include "body/main.h"

//////////////////////////////////////////////////////////////////////
////  BK::W32Font
//
namespace {

MAT2 const FontTransformMatrix = { {0, 1}, {0, 0}, {0, 0}, {0, 1} };

void buffercopy_glyph_to_film ( BK::Film::pixel_t * filmbuf,
                                const BK::Size2 & filmsize,
                                const char * glyphbuf,
                                const BK::Rect2 & glyphrect )
{
    using namespace BK;
    const Size2 glyphsize = glyphrect.get_size();
    
    const int rpitch = glyphsize.get_width() + ((4 - (glyphsize.get_width() & 3)) & 3); // DWORD alignment
    const int wpitch = filmsize.get_width();
    
    const char * rline = glyphbuf;
    Film::pixel_t * wline = filmbuf + glyphrect.get_left() + glyphrect.get_top() * wpitch;
    int x, y;
    
    for( y = glyphrect.get_height(); y > 0; --y ) {
        const char * rptr = rline;
        Film::pixel_t * wptr = wline;
        
        for( x = glyphrect.get_width(); x > 0; --x ) {
            // Windows ɂ 8 rbgmNtHg\zĂ̂ 64 Ŋ
            *wptr = (static_cast<int>(*rptr) * COLOR_ELEMENT_MAX) / 64;
            ++wptr;
            ++rptr;
        }
        rline += rpitch;
        wline += wpitch;
    }
}

inline UINT create_glyph_build_strbuf ( const char * str, char * buf )
{
    const unsigned char * & ustr = reinterpret_cast< const unsigned char * & >( str );
    
    buf[0] = str[0];
    if( (str[0] & 0x80) == 0 ) {
        buf[1] = '\0';
        return static_cast< unsigned int >(ustr[0]);
    } else {
        buf[1] = str[1];
        buf[2] = '\0';
        return (static_cast< unsigned int >(ustr[0]) << 8) | (static_cast< unsigned int >(ustr[1]));
    }
}

inline DWORD create_glyph_get_glyph_info ( HDC dc, UINT code, GLYPHMETRICS * gm )
{
    using namespace BK;
    const MAT2 & mat = FontTransformMatrix;
    DWORD reqsize;
    
    if( code == 0x20 ) {
        // 󔒕 (AEgCȂ)
        reqsize = 0;
    }
    else {
        // 擾\ȕ
        reqsize = GetGlyphOutline( dc, code, GGO_GRAY8_BITMAP, gm, 0, NULL, &mat );
        if( reqsize == 0 ) {
            // O𓊂󔒂ɕϊ
            //throw new windows_exception(" %04XH ̃rbg}bv擾Ɏs.", code);
            log_error(" %04XH ̃rbg}bv擾Ɏs.", code);
            reqsize = 0;
        }
    }
    return reqsize;
}

} /* namespace */

BK::W32Font::W32Font ( const BK::Font::info_t & info )
        : handle_device_context (NULL),
          handle_font (NULL),
          handle_font_backup (NULL),
          ptr_fontname_string (NULL)
{
    try {
        if( info.font_height <= 0 ) {
            throw new argument_exception("sȃtHg̍ (%d) ^܂.", info.font_height);
        }
        
        handle_device_context = CreateCompatibleDC( NULL );
        if( handle_device_context == NULL ) {
            throw new windows_exception("tHgŉʂƌ݊̂foCXReLXg̍쐬Ɏs.");
        }
        
        LOGFONT & f = obj_logical_font;
        
        ZeroMemory( &f, sizeof(LOGFONT) );
        f.lfHeight = info.font_height;
        f.lfWidth = 0;
        f.lfWeight = info.font_weight;
        f.lfItalic = (info.is_italic) ? TRUE : FALSE;
        f.lfUnderline = FALSE;
        f.lfStrikeOut = FALSE;
        f.lfCharSet = DEFAULT_CHARSET;
        f.lfOutPrecision = (info.is_raster_prior) ? OUT_RASTER_PRECIS : OUT_TT_PRECIS;
        f.lfClipPrecision = CLIP_DEFAULT_PRECIS;
        f.lfQuality = DEFAULT_QUALITY;
        f.lfPitchAndFamily = static_cast<BYTE>( ((info.is_proportional) ? VARIABLE_PITCH : FIXED_PITCH) | FF_DONTCARE );
        info.font_name.copy( f.lfFaceName, LF_FACESIZE - 1 );
        
        handle_font = CreateFontIndirect( &obj_logical_font );
        if( handle_font == NULL ) {
            throw new windows_exception("tHgu%sv̐Ɏs܂.", info.font_name.c_str());
        }
        
        handle_font_backup = reinterpret_cast<HFONT>( SelectObject( handle_device_context, handle_font ) );
        if( handle_font_backup == NULL ) {
            throw new windows_exception("tHgu%svfoCXReLXgɐݒł܂.", info.font_name.c_str());
        }
        
        if( GetTextMetrics( handle_device_context, &obj_metric ) == 0 ) {
            throw new windows_exception("tHgu%sṽgNX擾Ɏs.", info.font_name.c_str());
        }
        if( log_is_trace() ) {
            log_trace("TEXTMETRIC: tmHeight=%d, tmAscent=%d, tmDescent=%d",
                      obj_metric.tmHeight, obj_metric.tmAscent, obj_metric.tmDescent);
        }
        
        if( log_is_trace() ) {
            log_trace("Windows tHgu%s (TCY: %d)v𐶐܂.", info.font_name.c_str(), info.font_height);
        }
    } catch ( exception_base * ) {
        release();
        throw;
    }
}

BK::W32Font::~W32Font ()
{
    release();
}

void BK::W32Font::release ( void )
{
    if( handle_device_context != NULL ) {
        SelectObject( handle_device_context, handle_font_backup );
    }
    
    if( handle_font != NULL ) {
        DeleteObject( handle_font );
    }
    
    if( handle_device_context != NULL ) {
        DeleteDC( handle_device_context );
    }
    
    if( ptr_fontname_string != NULL ) {
        delete[] ptr_fontname_string;
    }
}

unsigned int BK::W32Font::get_charcode ( const char * str ) const
{
    const unsigned char * & ustr = reinterpret_cast< const unsigned char * & >( str );
    
    if( (str[0] & 0x80) == 0 ) {
        return static_cast< unsigned int >(ustr[0]);
    } else {
        return (static_cast< unsigned int >(ustr[0]) << 8) | (static_cast< unsigned int >(ustr[1]));
    }
}

BK::Glyph * BK::W32Font::create_glyph ( const char * str )
{
    typedef Film::pixel_t pixel_t;
    
    if( str == NULL ) {
        throw new argument_exception("Otŕ NULL w肳܂.");
    }
    
    // AEgCȂ (䕶) ͏o͕s\
    char strbuf [3];
    const UINT code = create_glyph_build_strbuf( str, strbuf );
    if( code < 0x20 ) {
        return NULL;
    }
    
    // AEgC̎擾
    GLYPHMETRICS gm;
    const HDC &  dc       = handle_device_context;
    const DWORD  reqsize  = create_glyph_get_glyph_info( dc, code, &gm );
    
    auto_ary < char > bufferobj ( reqsize );
    
    char * const buffer = bufferobj.get();
    {
        const MAT2 & mat = FontTransformMatrix;
        if( reqsize > 0 ) {
            // 擾\ȕ̏ꍇ̓C[WĂ
            if( GetGlyphOutline( dc, code, GGO_GRAY8_BITMAP, &gm, reqsize, buffer, &mat ) == GDI_ERROR ) {
                throw new windows_exception(" %04XH ̃rbg}bv擾Ɏs.", code);
            }
            #if 0
            if( log_is_trace() ) {
                log_trace("GLYPHMETRICS ( %04XH): BlackBox = %d~%d, Origin = (%d,%d)", code,
                          gm.gmBlackBoxX, gm.gmBlackBoxY, gm.gmptGlyphOrigin.x, gm.gmptGlyphOrigin.y);
            }
            #endif
        }
    }
    
    // tBp̃obt@pӂăC[WRs[
    const ABC           abc        = query_ABC( code );
    const Size2         size       = query_size( code );
    const unsigned int  pixelarea  = size.get_area();
    
    auto_ary < pixel_t > pixelsobj ( pixelarea );
    
    pixel_t * const pixels = pixelsobj.get();
    ZeroMemory( pixels, sizeof(pixel_t) * pixelarea );
    
    if( reqsize > 0 ) {
        const Point2  basepoint ( gm.gmptGlyphOrigin.x , obj_metric.tmAscent - gm.gmptGlyphOrigin.y );
        const Size2   blackbox ( gm.gmBlackBoxX, gm.gmBlackBoxY );
        
        buffercopy_glyph_to_film( pixels, size, buffer, Rect2(basepoint, blackbox) );
    }
    
    Film * const film = new Film( size, pixels );
    try {
        return new Glyph( strbuf, film, Glyph::ABC_t( abc.abcA, abc.abcB, abc.abcC ) );
    } catch ( exception_base * ) {
        delete film;
        throw;
    }
}

BK::Size2 BK::W32Font::query_size ( UINT code )
{
    INT width;
    
    if( GetCharWidth32( handle_device_context, code, code, &width ) == 0 ) {
        throw new windows_exception(" %04XH ̕擾Ɏs.", code);
    }
    #if 0
    if( log_is_trace() ) {
        log_trace("W32Font::create_glyph():  %04XH ̃TCY ... %dx%d", code, width, obj_metric.tmHeight);
    }
    #endif
    
    return Size2( width, obj_metric.tmHeight );
}

ABC BK::W32Font::query_ABC ( UINT code )
{
    ABCFLOAT abcfloat;
    
    if( GetCharABCWidthsFloat( handle_device_context, code, code, &abcfloat ) == 0 ) {
        throw new windows_exception(" %04XH  ABC 擾Ɏs.", code);
    }
    
    ABC abc;
    
    abc.abcA = static_cast< int >( abcfloat.abcfA );
    abc.abcB = static_cast< UINT >( abcfloat.abcfB );
    abc.abcC = static_cast< int >( abcfloat.abcfC );
    #if 0
    if( log_is_trace() ) {
        log_trace("ABC-info ( %04XH): A=%4d, B=%4u, C=%4d", code, abc.abcA, abc.abcB, abc.abcC);
    }
    #endif
    
    return abc;
}

BK::Size2 BK::W32Font::query_size ( const char * str )
{
    return query_size( get_charcode(str) );
}

const char * BK::W32Font::get_font_name ( void ) const
{
    const int BufferSize = 1024;
    
    if( ptr_fontname_string == NULL ) {
        ptr_fontname_string = new char [BufferSize];
        
        if( GetTextFace( handle_device_context, BufferSize, ptr_fontname_string ) == 0 ) {
            delete[] ptr_fontname_string;
            ptr_fontname_string = NULL;
            throw new windows_exception("tHg̎擾Ɏs܂.");
        }
    }
    
    return ptr_fontname_string;
}

//////////////////////////////////////////////////////////////////////
