#include "stdafx.h"
#include "body/main.h"

//////////////////////////////////////////////////////////////////////
////  BK::DirectX::InputDevice
//

BK::DirectX::InputDevice::InputDevice ( IDirectInput8 * dev, DIDEVICEINSTANCE const & inst, DIDATAFORMAT const & format )
        : obj_instance(inst),
          obj_device (NULL)
{
    W32ApplicationDevice * app = W32ApplicationDevice::get_instance();
    
    if( app == NULL ) {
        throw new logic_exception("AvP[VfoCX\zOɃCvbgfoCX̍\zs܂.");
    }
    
    IDirectInputDevice8 * device;
    GUID const & guid = obj_instance.guidInstance;
    HRESULT result;
    HWND wnd;
    
    result = dev->CreateDevice( guid, &device, NULL );
    if( result != DI_OK ) {
        throw new directx_exception(result, "GUID %s  DirectInput foCXɎs.",
                                    Win32::GUID_to_str(guid).c_str());
    }
    if( log_is_comment() ) {
        log_comment("GUID %s  DirectInput foCX (%s) 𐶐.", Win32::GUID_to_str(guid).c_str(), inst.tszProductName);
    }
    
    try {
        result = device->SetDataFormat( &format );
        if( result != DI_OK ) {
            throw new directx_exception(result, "DirectInput foCX (%s) ̃f[^`ݒɎs.",
                                        inst.tszProductName);
        }
        if( log_is_comment() ) {
            log_comment("foCX̃f[^`ݒ. (IDirectInputDevice8.SetDataFormat())");
        }
        
        wnd = app->get_window_handle();
        result = device->SetCooperativeLevel( wnd, DISCL_FOREGROUND | DISCL_NONEXCLUSIVE );
        if( result != DI_OK ) {
            throw new directx_exception(result, "DirectInput foCX (%s) ̋xݒɎs.",
                                        inst.tszProductName);
        }
        if( log_is_comment() ) {
            log_comment("foCX̋xݒ. (IDirectInputDevice8.SetCooperativeLevel(): hwnd=%p)", wnd);
        }
        
        result = device->Acquire();
        if( result != DI_OK ) {
            throw new directx_exception(result, "DirectInput foCX (%s) ւ̃ANZX擾Ɏs.",
                                        inst.tszProductName);
        }
        if( result == S_FALSE ) {
            log_error("ANZX擾ς݂ DirectInput foCX (%s) ɑ΂ăANZX̍Ď擾s܂.",
                      inst.tszProductName);
        } else {
            if( log_is_comment() ) {
                log_comment("DirectInput foCX (%s) ̃ANZX擾.", inst.tszProductName);
            }
        }
        obj_device = device;
    }
    catch ( exception_base * ) {
        Release( device );
        throw;
    }
}

BK::DirectX::InputDevice::~InputDevice ()
{
    if( obj_device != NULL ) {
        log_comment("DirectInput foCX (%s) ̃ANZX܂.", obj_instance.tszProductName);
        obj_device->Unacquire();
        Release( obj_device );
    }
}

void BK::DirectX::InputDevice::get_device_state ( DWORD size, LPVOID buf )
{
    HRESULT result;
    
    result = obj_device->GetDeviceState( size, buf );
    if( result != DI_OK ) {
        if( result != DIERR_INPUTLOST ) {
            throw new directx_exception(result, "DirectInput foCX (%s) ̏Ԏ擾Ɏs.",
                                        obj_instance.tszProductName);
        }
        result = obj_device->Acquire();
        if( result != DI_OK ) {
            throw new directx_exception(result, "DirectInput foCX (%s) ̃ANZXĎ擾Ɏs.",
                                        obj_instance.tszProductName);
        }
        log_info("DirectInput foCX (%s) Xg߃ANZX̍Ď擾s܂.");
        
        result = obj_device->GetDeviceState( size, buf );
        if( result != DI_OK ) {
            throw new directx_exception(result, "DirectInput foCX (%s) ̏Ԏ擾Ɏs.",
                                        obj_instance.tszProductName);
        }
    }
}

//////////////////////////////////////////////////////////////////////
////  BK::Dx8MouseInput
//

BK::Dx8MouseInput::Dx8MouseInput ( IDirectInput8 * dev, DIDEVICEINSTANCE const & inst )
        : InputDevice (dev, inst, c_dfDIMouse)
{
    ZeroMemory( &obj_state, sizeof(obj_state) );
}

BK::Dx8MouseInput::~Dx8MouseInput ()
{
}

const char * BK::Dx8MouseInput::get_name ( void ) const
{
    return get_device_instance().tszProductName;
}

void BK::Dx8MouseInput::update ( void )
{
    get_device_state( &obj_state );
}

BK::Point2 BK::Dx8MouseInput::get_delta ( void ) const
{
    return BK::Point2( static_cast<int>( obj_state.lX ), static_cast<int>( obj_state.lY ) );
}

void BK::Dx8MouseInput::get_button_states ( int * index, int start, int count ) const
{
    AbstractInput::get_button_states(index, start, count);
    switch( start ) {
    case 0:
        *index = (obj_state.rgbButtons[0] & 0x80) ? BUTSTAT_PRESSED : BUTSTAT_RELEASED;
        if( --count == 0 )
            break;
        ++index;
        
    case 1:
        *index = (obj_state.rgbButtons[1] & 0x80) ? BUTSTAT_PRESSED : BUTSTAT_RELEASED;
        if( --count == 0 )
            break;
        ++index;
        
    case 2:
        *index = (obj_state.lZ > 0) ? BUTSTAT_PRESSED : BUTSTAT_RELEASED;
        if( --count == 0 )
            break;
        ++index;
        
    case 3:
        *index = (obj_state.lZ < 0) ? BUTSTAT_PRESSED : BUTSTAT_RELEASED;
        if( --count == 0 )
            break;
        ++index;
        
    case 4:
        *index = (obj_state.rgbButtons[2] & 0x80) ? BUTSTAT_PRESSED : BUTSTAT_RELEASED;
        if( --count == 0 )
            break;
        ++index;
    }
}

//////////////////////////////////////////////////////////////////////
////  BK::Dx8KeyboardInput
//

BK::Dx8KeyboardInput::Dx8KeyboardInput ( IDirectInput8 * dev, DIDEVICEINSTANCE const & inst )
        : InputDevice (dev, inst, c_dfDIKeyboard)
{
    ZeroMemory( ary_state, sizeof(char) * 256 );
}

BK::Dx8KeyboardInput::~Dx8KeyboardInput ()
{
}

const char * BK::Dx8KeyboardInput::get_name ( void ) const
{
    return get_device_instance().tszProductName;
}

void BK::Dx8KeyboardInput::update ( void )
{
    get_device_state( sizeof(char) * 256, ary_state );
}

BK::Point2 BK::Dx8KeyboardInput::get_delta ( void ) const
{
    return BK::Point2( 0, 0 );
}

void BK::Dx8KeyboardInput::get_button_states ( int * index, int start, int count ) const
{
    int * wptr;
    const char * rptr;
    
    AbstractInput::get_button_states(index, start, count);
    
    wptr = index;
    rptr = &ary_state[ start ];
    for( ; count > 0; --count ) {
        *(wptr++) = (*(rptr++) & 0x80) ? BUTSTAT_PRESSED : BUTSTAT_RELEASED;
    }
}

//////////////////////////////////////////////////////////////////////
////  BK::Dx8GamepadInput
//

BK::Dx8GamepadInput::Dx8GamepadInput ( IDirectInput8 * dev, DIDEVICEINSTANCE const & inst )
        : InputDevice (dev, inst, c_dfDIJoystick),
          num_button_count (0)
{
    ZeroMemory( &obj_state, sizeof(obj_state) );
    
    DIDEVCAPS  caps;
    HRESULT    result;
    
    result = get_interface()->GetCapabilities( &caps );
    if( result != DI_OK ) {
        throw new directx_exception(result, "DirectInput foCX (%s) ̃f[^擾Ɏs.",
                                    inst.tszProductName);
    }
    num_button_count = caps.dwButtons;
}

BK::Dx8GamepadInput::~Dx8GamepadInput ()
{
}

const char * BK::Dx8GamepadInput::get_name ( void ) const
{
    return get_device_instance().tszProductName;
}

void BK::Dx8GamepadInput::update ( void )
{
    get_device_state( &obj_state );
}

BK::Point2 BK::Dx8GamepadInput::get_delta ( void ) const
{
    return BK::Point2( static_cast<int>( obj_state.lX ), static_cast<int>( obj_state.lY ) );
}

int BK::Dx8GamepadInput::count_button ( void ) const
{
    return num_button_count;
}

void BK::Dx8GamepadInput::get_button_states ( int * index, int start, int count ) const
{
    int * wptr;
    const BYTE * rptr;
    
    AbstractInput::get_button_states(index, start, count);
    
    wptr = index;
    rptr = &obj_state.rgbButtons[start];
    for( ; count > 0; --count ) {
        *(wptr++) = (*(rptr++) & 0x80) ? BUTSTAT_PRESSED : BUTSTAT_RELEASED;
    }
}

//////////////////////////////////////////////////////////////////////
////  BK::Dx8InputDevice
//

BK::Dx8InputDevice * BK::Dx8InputDevice::singleton_object = NULL;

BK::Dx8InputDevice::Dx8InputDevice ( void )
        : ptr_direct_input (NULL)
{
    if( W32ApplicationDevice::get_instance() == NULL ) {
        throw new logic_exception("AvP[VfoCX\zOɃCvbgfoCX\z܂.");
    }
    singleton_object = this;
}

BK::Dx8InputDevice::~Dx8InputDevice ()
{
    release_all();
}

void BK::Dx8InputDevice::release_all ( void )
{
    device_list_t::iterator i;
    
    for( i = ary_device.begin(); i != ary_device.end(); ++i ) {
        device_t & dev = *i;
        AbstractInput * vin = dev.input_object;
        
        if( vin != NULL && NIL_P(vin->get_ruby_object()) ) {
            delete vin;
        }
    }
    ary_device.clear();
    
    if( ptr_direct_input != NULL ) {
        log_comment("DirectInput IuWFNg܂.");
        DirectX::Release( ptr_direct_input );
        ptr_direct_input = NULL;
    }
}

void BK::Dx8InputDevice::_start ( void )
{
    //TraceScopeLog methodname ("BK::Dx8InputDevice::_start()");
    InputDevice::_start();
    
    try {
        W32ApplicationDevice * app = W32ApplicationDevice::get_instance();
        HRESULT result = DirectInput8Create(
            app->get_instance_handle(),
            DIRECTINPUT_VERSION,
            IID_IDirectInput8,
            reinterpret_cast<void**>(&ptr_direct_input),
            NULL );
        if( result != DI_OK || ptr_direct_input == NULL ) {
            throw new directx_exception(result, "DirectInput IuWFNg̐Ɏs܂");
        }
        log_comment("DirectInput IuWFNg <%p> 𐶐܂.", ptr_direct_input);
        
        DirectX::inputdevinst_list_t::iterator  i;
        DirectX::inputdevinst_list_t            instlist;
        
        DirectX::enum_input_devices( ptr_direct_input, &instlist );
        if( log_is_comment() ) {
            {
                std::string buffer = "FꂽCvbgfoCẌꗗ:\n";
                for( i = instlist.begin(); i != instlist.end(); ++i ) {
                    if( i != instlist.begin() ) {
                        buffer.append( "\n" );
                    }
                    buffer.append( DirectX::inspect_input_device_instance(*i) );
                }
                log_write_comment(buffer);
            }
        }
        for( i = instlist.begin(); i != instlist.end(); ++i ) {
            ary_device.push_back( device_t(*i) );
        }
        
    } catch ( exception_base * ) {
        release_all();
        throw;
    }
}

void BK::Dx8InputDevice::_stop ( void )
{
    //TraceScopeLog methodname ("BK::Dx8InputDevice::_stop()");
    release_all();
    InputDevice::_stop();
}

void BK::Dx8InputDevice::ruby_gc_mark ( void )
{
    InputDevice::ruby_gc_mark();
    std::for_each( ary_device.begin(), ary_device.end(), RubyObjectMarker() );
}

BK::Dx8InputDevice::device_t * BK::Dx8InputDevice::device_at ( int index )
{
    device_list_t::iterator i;
    
    if( index < 0 || index >= count_device() ) {
        throw new out_of_range_exception("w肳ꂽ̓foCXԍ (%d) ͔͈͊Oł. (0...%d)", index, count_device());
    }
    
    i = ary_device.begin();
    std::advance( i, index );
    
    return &(*i);
}

BK::Dx8InputDevice::device_t const * BK::Dx8InputDevice::device_at ( int index ) const
{
    device_list_t::const_iterator i;
    
    if( index < 0 || index >= count_device() ) {
        throw new out_of_range_exception("w肳ꂽ̓foCXԍ (%d) ͔͈͊Oł. (0...%d)", index, count_device());
    }
    
    i = ary_device.begin();
    std::advance( i, index );
    
    return &(*i);
}

int BK::Dx8InputDevice::count_device ( void ) const
{
    return ary_device.size();
}

std::string BK::Dx8InputDevice::get_device_name ( int index ) const
{
    char buf [ MAX_PATH * 3 ];
    device_t const * dev = device_at(index);
    DIDEVICEINSTANCE const & inst = dev->device_instance;
    
    sprintf(buf, "%s (%s)", inst.tszProductName, inst.tszInstanceName);
    
    return std::string( buf );
}

BK::AbstractInput::type_t BK::Dx8InputDevice::dwDevType_to_type ( DWORD devtype )
{
    devtype = GET_DIDEVICE_TYPE(devtype);
    
    switch( devtype ) {
    case DI8DEVTYPE_GAMEPAD:
    case DI8DEVTYPE_JOYSTICK:
        return AbstractInput::DEV_GAMEPAD;
        
    case DI8DEVTYPE_KEYBOARD:
        return AbstractInput::DEV_KEYBOARD;
        
    case DI8DEVTYPE_MOUSE:
        return AbstractInput::DEV_MOUSE;
    }
    
    return AbstractInput::DEV_UNKNOWN;
}

BK::AbstractInput::type_t BK::Dx8InputDevice::get_device_type ( int index ) const
{
    device_t const * dev = device_at(index);
    
    return dwDevType_to_type( dev->device_instance.dwDevType );
}

BK::AbstractInput * BK::Dx8InputDevice::setup_device ( int index )
{
    W32ApplicationDevice * app = W32ApplicationDevice::get_instance();
    
    if( app == NULL ) {
        throw new logic_exception("AvP[VfoCX\zOɓ̓foCX̍\zs܂.");
    }
    
    device_t * dev = device_at(index);
    DIDEVICEINSTANCE const & inst = dev->device_instance;
    AbstractInput * driver;
    
    driver = dev->input_object;
    if( driver != NULL ) {
        return driver;
    }
    
    log_comment("̓foCX (%s) ̃hCo\z܂.", inst.tszProductName);
    switch( dwDevType_to_type(inst.dwDevType) ) {
    case AbstractInput::DEV_GAMEPAD:
        driver = new Dx8MouseInput( ptr_direct_input, inst );
        break;
        
    case AbstractInput::DEV_KEYBOARD:
        driver = new Dx8KeyboardInput( ptr_direct_input, inst );
        break;
        
    case AbstractInput::DEV_MOUSE:
        driver = new Dx8MouseInput( ptr_direct_input, inst );
        break;
    }
    if( driver == NULL ) {
        throw new argument_exception("w肳ꂽfoCX (%s) ɑΉhCo܂.", inst.tszProductName);
    }
    
    dev->input_object = driver;
    return driver;
}

void BK::Dx8InputDevice::release_device ( int index )
{
    throw new unsupported_exception("Dx8InputDevice::release_device() ͖ł.");
}

//////////////////////////////////////////////////////////////////////
