#include "stdafx.h"
#include "boot/main.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  t@C[JȃO[oϐ
//
namespace {

static bool flag_ruby_closed = false;

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  t@C[JȊ֐
//
namespace {

void debug_msgbox ( bool flag, const char * format, ... )
{
    if( !flag ) {
        return;
    }
    
    va_list arg;
    
    va_start( arg, format );
    std::string str = BK::strvformat( format, arg );
    va_end( arg );
    
    MessageBox(NULL, str.c_str(), "u[g", MB_OK | MB_ICONINFORMATION );
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  O[o֐
//

void Init_bk ( void )
{
    // XL[}W[̏
    // RubyNX̐錾Agݍ݃W[̏s܂B
    BK::Init_schema();
    
    // {fBW[̏
    // ˑȃW[̏s܂B
    // (ł͎gpĂȂ̂ŃRgAEg)
    //BK::Init_body();
}

void Exit_bk ( void )
{
    // {fBW[̏I
    // ˑȃW[̒~s܂B
    // (ł͎gpĂȂ̂ŃRgAEg)
    //BK::Exit_body();
    
    // XL[}W[̏I
    // gݍ݃W[̒~s܂B
    BK::Exit_schema();
}

////////////////////////////////////////
namespace {

VALUE get_backtrace ( VALUE errinfo )
{
    VALUE volatile backtraceobj = rb_funcall( ruby_errinfo, rb_intern("backtrace"), 0 );
    VALUE volatile aryobj;
    
    aryobj = rb_check_array_type( backtraceobj );
    if( NIL_P(aryobj) ) {
        return rb_String( backtraceobj );
    }
    
    VALUE volatile lineary = rb_ary_new();
    int i;
    
    for( i = 0; i < xrb_ary_length(aryobj); ++i ) {
        VALUE volatile strobj = rb_check_string_type( xrb_ary_at(aryobj, i) );
        
        if( !NIL_P(strobj) ) {
            VALUE volatile spcobj = rb_str_new2("   at ");
            
            rb_str_append( spcobj, strobj );
            rb_ary_push( lineary, spcobj );
        }
    }
    
    return rb_String( rb_funcall(lineary, rb_intern("join"), 1, rb_str_new2("\n")) );
}

void on_ruby_exception ( VALUE errinfo )
{
    std::string title = BK::strformat("RubyO: %s", rb_obj_classname(ruby_errinfo));
    std::string errstr;
    VALUE volatile errstrobj = errinfo;
    VALUE volatile btrace;
    const char * errmes;
    int state;
    
    errmes = StringValuePtr( errstrobj );
    btrace = rb_protect( get_backtrace, errinfo, &state );
    errstr.append( errmes );
    if( state == 0 ) {
        if( TYPE(btrace) == T_STRING ) {
            errstr.append("\n");
            errstr.append( xrb_str_ptr(btrace) );
        }
    }
    
    MessageBox(NULL, errstr.c_str(), title.c_str(), MB_OK );
    BK::log_error("<<Ruby O>> [%s] %s", rb_obj_classname(ruby_errinfo), errstr.c_str());
}

void on_cpp_exception ( BK::exception_base * e )
{
    std::string title = BK::strformat("C++O: %s", e->get_type());
    
    MessageBox(NULL, e->what(), title.c_str(), MB_OK | MB_ICONERROR );
    BK::log_error("<<C++ O>> [%s] %s", e->get_type(), e->what());
}

} /* namespace */
////////////////////////////////////////
// gݍ݃Cȕ

extern "C" {

void Init_win32ole();

} /* extern "C" */

void Init_ruby_ext ( void )
{
    Init_win32ole();
}

////////////////////////////////////////
namespace {

int xrb_init ( int argc, char** argv, char** envp )
{
    NtInitialize(&argc, &argv);
    
    ruby_init();
    
    debug_msgbox(false, "Ruby C^v^̏܂.");
    
    rb_set_kcode("sjis"); // ftHgSJIS
    
    ruby_options(argc, argv);
    
    debug_msgbox(false, "R}hCIvV̉͂܂.");
    
    if (ruby_nerrs > 0) {
        return ruby_nerrs;
    }
    
    Init_ruby_ext();
    
    debug_msgbox(false, "gݍ݊gW[̏܂.");
    
    Init_bk();
    
    debug_msgbox(false, "BK W[̏܂.");
    
    return 0;
}

int xrb_boot ( void )
{
    // from ruby_run() [eval.c] //
    int state;
    
    try {
        state = ruby_exec();
        
        if( state ) {
            on_ruby_exception(ruby_errinfo);
        }
    } catch ( BK::exception_base * e ) {
        on_cpp_exception(e);
        delete e;
    }
    
    return state;
}

} /* namespace */
////////////////////////////////////////

int main ( int argc, char** argv, char** envp )
{
    using namespace BK;
    int state;
    
    debug_msgbox(false, "Black Kit N܂.");
    
    state = xrb_init( argc, argv, envp );
    if( state > 0 ) {
        MessageBox(NULL, strformat("Ɏs. (state=%d)", state).c_str(), "u[gG[", MB_OK | MB_ICONERROR );
        return state;
    }
    
    try {
        const char * const fname = STARTUP_LOG_OUTPUT_FILENAME;
        if( fname != NULL ) {
            LogDevice * dev = LogDevice::get_instance();
            std::ostream * out = new std::ofstream ( fname );
            
            dev->start();
            dev->set_logger_stream(out);
        }
    }
    catch ( exception_base * e ) {
        on_cpp_exception(e);
        delete e;
    }
    
    state = xrb_boot();
    log_info("Ruby C^v^̎sI܂. (state=%d)", state);
    
    try {
        state = ruby_cleanup(state);
    } catch ( exception_base * e ) {
        on_cpp_exception(e);
        delete e;
    }
    flag_ruby_closed = true;
    
    log_info("Ruby C^v^Vbg_E܂. (state=%d)", state);
    Exit_bk();
    
    return state;
}

//////////////////////////////////////////////////////////////////////
////  O[o֐
//

bool BK::is_rubysys_closed ( void )
{
    return flag_ruby_closed;
}

//////////////////////////////////////////////////////////////////////
