#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::CharacterGear
//
namespace {
extern "C" {

/// VLN^[MA𐶐
VALUE CharacterGear_new ( VALUE klass )
{
    using namespace BK;
    VALUE volatile obj = Qnil;
    CharacterGear * ptr;
    
    try {
        ptr = new CharacterGear();
        if( ptr == NULL ) {
            throw_no_memory_for_object("CharacterGear");
        }
        obj = ptr->ruby_wrap();
    }
    catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    rb_obj_call_init( obj, 0, NULL );
    
    try {
        ApplicationDevice::global_add_gear( *ptr );
        if( log_is_trace() ) {
            log_trace("LN^[MA <%s:%p> 𐶐܂.", typeid(*ptr).name(), ptr);
        }
    }
    catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return obj;
}

/// LN^[̏s
VALUE CharacterGear_initialize ( VALUE self )
{
    return self;
}

/// LN^[𐶐
VALUE CharacterGear_generate ( VALUE self )
{
    using namespace BK;
    CharacterGear * const ptr = CharacterGear::value_to_data( self );
    VALUE volatile ret = Qnil;
    
    try {
        Character * chr = new Character();
        
        if( chr == NULL ) {
            throw_no_memory_for_object("Character");
        }
        
        ret = chr->ruby_wrap();
        
        ptr->add( *chr );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return ret;
}

/// nꂽXvCgtB[hLN^[𐶐
VALUE CharacterGear_generate_sprite ( VALUE self, VALUE fieldobj )
{
    using namespace BK;
    CharacterGear * const ptr = CharacterGear::value_to_data( self );
    VALUE volatile ret = Qnil;
    
    if( !SpriteField::is_kind_of(fieldobj) ) {
        rb_raise(rb_eArgError, "XvCgLN^[̐ SpriteField ȊO (%s) ͎ws.",
                 rb_obj_classname(fieldobj));
    }
    
    try {
        SpriteField * const field = SpriteField::value_to_data( fieldobj );
        SpriteCharacter * const chr = new SpriteCharacter( *field );
        
        if( chr == NULL ) {
            throw_no_memory_for_object("SpriteCharacter");
        }
        
        ret = chr->ruby_wrap();
        
        ptr->add( *chr );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return ret;
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::CharacterGear
//

VALUE BK::CharacterGear::ruby_class_object = Qnil;

BK::CharacterGear::CharacterGear ( void )
        : ary_character  (NULL),
          num_size       (0),
          num_capacity   (CHARACTER_GEAR_INITIAL_SIZE)
{
    this->ary_character = new Character * [ num_capacity ];
}

BK::CharacterGear::~CharacterGear ()
{
    // ary_character ۗLĂIuWFNg͊O^ꂽ̂Ȃ̂ŁA
    // ɏ_B
    delete[] this->ary_character;
}

VALUE BK::CharacterGear::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "CharacterGear", parent_t::get_ruby_class() );
    
    xrb_define_0arg_singleton_method( c, "new", CharacterGear_new );
    
    xrb_define_0arg_method( c, "initialize", CharacterGear_initialize );
    xrb_define_0arg_method( c, "generate", CharacterGear_generate );
    xrb_define_1arg_method( c, "generate_sprite", CharacterGear_generate_sprite );
    
    return ruby_class_object;
}

void BK::CharacterGear::ruby_gc_mark ( void )
{
    parent_t::ruby_gc_mark();
    
    Character * const * const ary = this->ary_character;
    const unsigned int s = this->num_size;
    unsigned int i;
    
    for( i = 0; i < s; ++i ) {
        Character * const chr = ary[i];
        chr->gc_mark_self();
    }
}

void BK::CharacterGear::on_ruby_wrap ( void )
{
    parent_t::on_ruby_wrap();
    
    Character * const * const ary = this->ary_character;
    const unsigned int s = this->num_size;
    unsigned int i;
    
    for( i = 0; i < s; ++i ) {
        Character * const chr = ary[i];
        chr->ruby_wrap();
    }
}

void BK::CharacterGear::execute ( void )
{
    Character * * const ary = this->ary_character;
    unsigned int s = this->num_size;
    unsigned int i = 0;
    
    while( i < s ) {
        Character * const chr = ary[i];
        
        if( chr->is_terminated() ) {
            --s;
            if( i < s ) {
                ary[i] = ary[s];
            }
        } else {
            if( chr->is_active() ) {
                chr->execute();
            }
            ++i;
        }
    }
    this->num_size = s;
}

void BK::CharacterGear::add ( BK::Character & chr )
{
    if( this->num_size >= this->num_capacity ) {
        const unsigned int newcap = this->num_capacity * 2;
        Character * * newbuf = new Character * [ newcap ];
        
        memcpy( newbuf, this->ary_character, sizeof(Character *) * this->num_capacity );
        delete[] this->ary_character;
        this->ary_character = newbuf;
        this->num_capacity = newcap;
    }
    this->ary_character[ (this->num_size)++ ] = &chr;
}

BK::Character * BK::CharacterGear::erase ( BK::Character * ptr )
{
    log_error("Ȃ\bh CharacterGear.erase() gp܂.");
    
    Character * * const ary = this->ary_character;
    unsigned int s = this->num_size;
    unsigned int i;
    
    for( i = 0; i < s; ++i ) {
        if( ary[i] == ptr ) {
            ary[i] = ary[--s];
            this->num_size = s;
            return ptr;
        }
    }
    throw new logic_exception("LN^[MA <%p> ɑ݂ȂLN^[ <%p> ̍폜s܂.", this, ptr);
    return NULL; // dummy
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_application_gear_CharacterGear ( void )
{
    CharacterGear::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_application_gear_CharacterGear ( void )
{
}

//////////////////////////////////////////////////////////////////////
