#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::Gear
//
namespace {
extern "C" {

/// Gear 
void Gear_free ( BK::RubyObject * ptr )
{
    using namespace BK;
    //TraceScopeLog scopelog ("Gear_free()");
    Gear * const gear = dynamic_cast< Gear * >( ptr );
    
    if( gear == NULL ) {
        if( ptr == NULL ) {
            log_error("[Gear_free()] Gear_free()  NULL n܂.");
        } else {
            log_error("[Gear_free()] Gear_free()  <%s:%p> n܂.", typeid(*ptr).name(), ptr);
        }
        return;
    }
    
    if( !gear->is_terminated() ) {
        gear->terminate();
    }
    
    if( gear != NULL ) {
        ApplicationDevice::global_erase_gear( gear );
    } else {
        log_error("[Gear_free()] Gear ȊÕIuWFNg Gear_free() gpĂ܂.");
    }
    if( log_is_debug() && ptr->is_ruby_gc_log_enabled() ) {
        log_debug("Ruby GC  %s <%p> ܂.", typeid(*ptr).name(), ptr);
    }
    delete ptr;
}

/// MA ID 擾
VALUE Gear_get_id ( VALUE self )
{
    using namespace BK;
    Gear * const gear = Gear::value_to_data( self );
    
    return rb_uint_new(gear->get_gear_id());
}

/// MA ID ݒ肷
VALUE Gear_set_id ( VALUE self, VALUE numobj )
{
    using namespace BK;
    rb_check_frozen( self );
    
    Gear * const gear = Gear::value_to_data( self );
    const unsigned int id = NUM2UINT(numobj);
    
    try {
        gear->set_gear_id( id );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return numobj;
}

/// MÅԂ擾
VALUE Gear_is_active ( VALUE self )
{
    using namespace BK;
    Gear * const ptr = Gear::value_to_data( self );
    
    return OnOffObject_is_active( ptr );
}

/// MAԂɐݒ肷
VALUE Gear_activate ( VALUE self )
{
    using namespace BK;
    rb_check_frozen( self );
    
    Gear * const ptr = Gear::value_to_data( self );
    
    OnOffObject_activate( ptr );
    
    return self;
}

/// MA񊈐Ԃɐݒ肷
VALUE Gear_deactivate ( VALUE self )
{
    using namespace BK;
    rb_check_frozen( self );
    
    Gear * const ptr = Gear::value_to_data( self );
    
    OnOffObject_deactivate( ptr );
    
    return self;
}

/// MA~Ԃɐݒ肷
VALUE Gear_terminate ( VALUE self )
{
    using namespace BK;
    rb_check_frozen( self );
    
    Gear * const ptr = Gear::value_to_data( self );
    
    OnOffObject_terminate( ptr );
    
    return self;
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::Gear
//

VALUE BK::Gear::ruby_class_object = Qnil;

BK::Gear::Gear ( void )
        : OnOffObject     (STAT_ACTIVE),
          gear_id         (0),
          num_last_time   (0),
          num_total_time  (0),
          num_active_time (0)
{
    // AvP[VfoCXւ̃MAo^́AeTuNXōsKv܂B
}

BK::Gear::~Gear ()
{
}

VALUE BK::Gear::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "Gear", BK::RubyObject::get_ruby_class() );
    
    xrb_define_0arg_method( c, "id", Gear_get_id );
    xrb_define_1arg_method( c, "id=", Gear_set_id );
    xrb_define_0arg_method( c, "active?", Gear_is_active );
    xrb_define_0arg_method( c, "activate", Gear_activate );
    xrb_define_0arg_method( c, "deactivate", Gear_deactivate );
    xrb_define_0arg_method( c, "terminate", Gear_terminate );
    
    return ruby_class_object;
}

void * BK::Gear::get_ruby_free_function ( void )
{
    return Gear_free;
}

// ~ԂɂȂAvP[V폜Ă̂ł̃NX邱Ƃ͂Ȃɂ
//void BK::Gear::on_activate ( BK::OnOffObject::stat_t status )
//{
//    OnOffObject::on_activate( status );
//}

void BK::Gear::set_gear_id ( unsigned int id )
{
    BK::ApplicationDevice * const app = BK::ApplicationDevice::get_instance();
    
    if( app == NULL ) {
        throw new logic_exception("AvP[VfoCX\zɃMA ID ͐ݒs.");
    }
    
    if( id == gear_id ) {
        if( log_is_trace() ) {
            log_trace("MA <%s:%p>  ID  %u ɍĐݒ肳܂.", typeid(*this).name(), this, id);
        }
    }
    else {
        if( log_is_trace() ) {
            log_trace("MA <%s:%p>  ID  %u  %u ɕύX܂.", typeid(*this).name(), this, gear_id, id);
        }
        
        // MA ID ςʒmKv
        ApplicationDevice::global_erase_gear(this);
        
        this->gear_id = id;
        
        ApplicationDevice::global_add_gear(*this);
    }
}

void BK::Gear::notify_time ( BK::FrameTimer::time_t t )
{
    // ߂Ă notify ͎Zbĝ
    if( this->num_last_time == 0 && this->num_total_time == 0 ) {
        this->num_last_time = t;
        return;
    }
    
    const FrameTimer::time_t delta = FrameTimer::get_time_distance( this->num_last_time, t );
    
    this->num_total_time += delta;
    if( this->is_active() ) {
        this->num_active_time += delta;
    }
    this->num_last_time = t;
}

void BK::Gear::set_active_time ( BK::FrameTimer::time_t t )
{
    this->num_active_time = t;
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_application_Gear ( void )
{
    Gear::get_ruby_class();
    
    Init_application_gear_GearArray();
    Init_application_gear_ManualGear();
    Init_application_gear_TimerGear();
    Init_application_gear_CharacterGear();
    Init_application_gear_FadeInGear();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_application_Gear ( void )
{
    Exit_application_gear_FadeInGear();
    Exit_application_gear_ManualGear();
    Exit_application_gear_TimerGear();
    Exit_application_gear_CharacterGear();
    Exit_application_gear_GearArray();
}

//////////////////////////////////////////////////////////////////////
