#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  t@C[JȒ萔
//
namespace {


/// MAXgۗL郊Xg̏TCY
const unsigned int GEAR_ARRAY_INITIAL_SIZE = 128;


} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::GearArray
//

BK::GearArray::GearArray ( void )
        : OnOffObject   (STAT_ACTIVE),
          ary_gear      (new Gear * [ GEAR_ARRAY_INITIAL_SIZE ]),
          num_size      (0),
          num_capacity  (GEAR_ARRAY_INITIAL_SIZE)
{
}

BK::GearArray::~GearArray ()
{
    delete[] ary_gear;
}

void BK::GearArray::ruby_gc_mark ( void )
{
    unsigned int i;
    
    for( i = 0; i < num_size; ++i ) {
        ary_gear[i]->gc_mark_self();
    }
}

void BK::GearArray::broadcast ( BK::FrameTimer::time_t t )
{
    unsigned int i = 0;
    unsigned int s = num_size;
    
    while( i < s ) {
        Gear * const gear = ary_gear[i];
        
        if( gear->is_terminated() ) {
            --s;
            if( i < s ) {
                #if 1
                if( log_is_trace() ) {
                    log_trace("GearArray <%p> ~MA <%s:%p> 폜܂.",
                              this, typeid(*ary_gear[i]).name(), ary_gear[i]);
                }
                #endif
                ary_gear[i] = ary_gear[s];
            }
        } else {
            gear->notify_time( t );
            if( gear->is_active() ) {
                gear->execute();
            }
            ++i;
        }
    }
    num_size = s;
}

void BK::GearArray::add_gear ( BK::Gear & gear )
{
    #if 1
    if( log_is_trace() ) {
        log_trace("GearArray <%p> ɃMA <%s:%p> ǉ܂.", this, typeid(gear).name(), &gear);
    }
    #endif
    
    if( num_size >= num_capacity ) {
        const unsigned int newcap = num_capacity * 2;
        Gear * * const newbuf = new Gear * [ newcap ];
        
        memcpy( newbuf, ary_gear, sizeof(Gear *) * num_capacity );
        delete[] ary_gear;
        ary_gear = newbuf;
        num_capacity = newcap;
    }
    ary_gear[ num_size++ ] = &gear;
}

BK::Gear * BK::GearArray::erase_gear ( BK::Gear * gear )
{
    for( unsigned int i = 0; i < num_size; ++i ) {
        if( ary_gear[i] == gear ) {
            --num_size;
            ary_gear[i] = ary_gear[num_size];
            #if 1
            if( log_is_trace() ) {
                log_trace("GearArray <%p> MA <%s:%p> 폜܂.", this, typeid(*gear).name(), gear);
            }
            #endif
            return gear;
        }
    }
    return NULL;
}

bool BK::GearArray::is_include ( Gear * gear ) const
{
    for( unsigned int i = 0; i < num_size; ++i ) {
        if( ary_gear[i] == gear ) {
            return true;
        }
    }
    return false;
}

//////////////////////////////////////////////////////////////////////
////  BK::GearStorage
//

BK::GearStorage::GearStorage ( void )
{
}

BK::GearStorage::~GearStorage ()
{
    clear();
}

void BK::GearStorage::add_gear ( BK::Gear & gear )
{
    const unsigned int id = gear.get_gear_id();
    const list_t::iterator e = ary_list.end();
    list_t::iterator itr;
    GearArray * target = NULL;
    
    for( itr = ary_list.begin(); target == NULL && itr != e; ++itr ) {
        if( itr->first == id ) {
            target = itr->second;
        }
    }
    if( target == NULL ) {
        target = new GearArray ();
        ary_list.push_back( element_t(id, target) );
        std::sort( ary_list.begin(), ary_list.end() );
    }
    target->add_gear( gear );
}

BK::Gear * BK::GearStorage::erase_gear ( BK::Gear * gear )
{
    if( gear == NULL ) {
        return NULL;
    }
    
    const unsigned int id = gear->get_gear_id();
    const list_t::iterator itr = std::find( ary_list.begin(), ary_list.end(), id );
    
    if( itr == ary_list.end() ) {
        return NULL;
    }
    
    GearArray * const gears = itr->second;
    
    if( gears->erase_gear( gear ) == NULL ) {
        return NULL;
    }
    if( gears->size() == 0 ) {
        ary_list.erase( itr );
        delete gears;
    }
    
    return gear;
}

void BK::GearStorage::clear ( void )
{
    const list_t::iterator e = ary_list.end();
    list_t::iterator itr;
    
    for( itr = ary_list.begin(); itr != e; ++itr ) {
        delete itr->second;
    }
    ary_list.clear();
}

void BK::GearStorage::broadcast ( BK::FrameTimer::time_t t )
{
    const list_t::iterator e = ary_list.end();
    list_t::iterator itr;
    
    for( itr = ary_list.begin(); itr != e; ++itr ) {
        GearArray * const gears = itr->second;
        
        if( gears->is_active() ) {
            gears->broadcast( t );
        }
    }
}

void BK::GearStorage::ruby_gc_mark ( void )
{
    const list_t::iterator e = ary_list.end();
    list_t::iterator itr;
    
    for( itr = ary_list.begin(); itr != e; ++itr ) {
        itr->second->ruby_gc_mark();
    }
}

unsigned int BK::GearStorage::size ( void ) const
{
    const list_t::const_iterator e = ary_list.end();
    list_t::const_iterator itr;
    unsigned int size = 0;
    
    for( itr = ary_list.begin(); itr != e; ++itr ) {
        size += itr->second->size();
    }
    
    return size;
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_application_gear_GearArray ( void )
{
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_application_gear_GearArray ( void )
{
}

//////////////////////////////////////////////////////////////////////
