#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: LinearAlphaScript
//
namespace {
extern "C" {

/// XNvg̐
VALUE LinearAlphaScript_new ( VALUE klass )
{
    using namespace BK;
    VALUE volatile obj = Qnil;
    LinearAlphaScript * ptr;
    
    try {
        ptr = new LinearAlphaScript ();
        if( ptr == NULL ) {
            throw_no_memory_for_object("LinearAlphaScript");
        }
        obj = ptr->ruby_wrap();
    }
    catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    rb_obj_call_init( obj, 0, NULL );
    
    return obj;
}

/// XNvg̏
VALUE LinearAlphaScript_initialize ( VALUE self )
{
    return self;
}

/// l̎擾
VALUE LinearAlphaScript_get_range ( VALUE self )
{
    using namespace BK;
    LinearAlphaScript * const ptr = LinearAlphaScript::value_to_data( self );
    VALUE volatile ret;
    
    // 0: I[܂
    ret = rb_range_new( rb_fix_new(ptr->get_begin()),
                        rb_fix_new(ptr->get_end()),
                        0 );
    
    return ret;
}

/// l̐ݒ
VALUE LinearAlphaScript_set_range ( VALUE self, VALUE arg )
{
    using namespace BK;
    LinearAlphaScript * const ptr = LinearAlphaScript::value_to_data( self );
    VALUE volatile v0;
    VALUE volatile v1;
    bool include_end = true;
    
    if( xrb_is_kind_of(arg, rb_cRange) ) {
        v0 = rb_funcall( arg, rb_intern("first"), 0 );
        v1 = rb_funcall( arg, rb_intern("last"), 0 );
        if( RTEST(rb_funcall(arg, rb_intern("exclude_end?"), 0)) ) {
            include_end = false;
        }
    }
    else if( xrb_is_kind_of(arg, rb_cArray) ) {
        if( xrb_ary_length(arg) != 2 ) {
            rb_raise(rb_eArgError, "LinearAlphaScript#range= ɂ͒ 2 ȊO (%d) ̔z͎wł܂.", xrb_ary_length(arg));
        }
        v0 = xrb_ary_at(arg, 0);
        v1 = xrb_ary_at(arg, 1);
    }
    else {
        rb_raise(rb_eArgError, "LinearAlphaScript#range= ɂ %s ݒł܂.", rb_obj_classname(arg));
    }
    
    unsigned int num0;
    unsigned int num1;
    
    num0 = NUM2UINT( v0 );
    num1 = NUM2UINT( v1 );
    if( !include_end ) {
        if( num0 < num1 ) {
            num1 -= 1;
        } else {
            num1 += 1;
        }
    }
    
    try {
        ptr->set_begin( num0 );
        ptr->set_end( num1 );
    }
    catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return arg;
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::LinearAlphaScript::Agent
//

BK::LinearAlphaScript::Agent::Agent ( BK::LinearAlphaScript & script )
        : agent_parent_t (script),
          ptr_script (&script),
          ptr_character (NULL)
{
}

BK::LinearAlphaScript::Agent::~Agent ()
{
}

bool BK::LinearAlphaScript::Agent::connect_character ( BK::Character * chr )
{
    if( !agent_parent_t::connect_character( chr ) ) {
        return false;
    }
    
    ptr_character = dynamic_cast< SpriteCharacter * >( chr );
    if( ptr_character == NULL ) {
        return false;
    }
    
    return true;
}

void BK::LinearAlphaScript::Agent::execute ( void )
{
    const AlphaColor * const color = ptr_character->get_color();
    AlphaColor c;
    
    if( color == NULL ) {
        c.set_alpha_color( 0xff, 0xff, 0xff, 0 );
    } else {
        c = *color;
    }
    
    const unsigned int len = ptr_script->get_length();
    unsigned int cur = get_current();
    
    if( cur <= len ) {
        c.set_alpha( (ptr_script->get_begin() * (len - cur) + ptr_script->get_end() * cur) / len );
        ptr_character->set_color( &c );
        cur += 1;
        set_current( cur );
    }
    
    agent_parent_t::execute();
    
    if( cur > len ) {
        terminate();
    }
}

//////////////////////////////////////////////////////////////////////
////  BK::LinearAlphaScript
//

VALUE BK::LinearAlphaScript::ruby_class_object = Qnil;

BK::LinearAlphaScript::LinearAlphaScript ( void )
        : num_begin  (0x00),
          num_end    (0xff)
{
}

BK::LinearAlphaScript::~LinearAlphaScript ()
{
}

VALUE BK::LinearAlphaScript::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "LinearAlphaScript", parent_t::get_ruby_class() );
    
    xrb_define_0arg_singleton_method( c, "new", LinearAlphaScript_new );
    
    xrb_define_0arg_method( c, "initialize", LinearAlphaScript_initialize );
    xrb_define_0arg_method( c, "range", LinearAlphaScript_get_range );
    xrb_define_1arg_method( c, "range=", LinearAlphaScript_set_range );
    
    return ruby_class_object;
}

BK::CharacterScript::Agent * BK::LinearAlphaScript::create_agent ( void )
{
    return new Agent( *this );
}

void BK::LinearAlphaScript::set_begin ( unsigned int n )
{
    if( n > COLOR_ELEMENT_MAX ) {
        throw new argument_exception(
            "LinearAlphaScript ɔ͈͊O (0 ` %u) ̊JnAt@l %u w肳܂.", COLOR_ELEMENT_MAX, n);
    }
    num_begin = n;
}

void BK::LinearAlphaScript::set_end ( unsigned int n )
{
    if( n > COLOR_ELEMENT_MAX ) {
        throw new argument_exception(
            "LinearAlphaScript ɔ͈͊O (0 ` %u) ̏IAt@l %u w肳܂.", COLOR_ELEMENT_MAX, n);
    }
    num_end = n;
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_application_character_LinearAlphaScript ( void )
{
    LinearAlphaScript::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_application_character_LinearAlphaScript ( void )
{
}

//////////////////////////////////////////////////////////////////////
