#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::SpriteCharacter
//
namespace {
extern "C" {

/// XvCgLN^[̒S_擾
VALUE SpriteCharacter_get_center ( VALUE self )
{
    BK::SpriteCharacter * const ptr = BK::SpriteCharacter::value_to_data( self );
    
    return ptr->get_center().create_ruby_object();
}

/// XvCgLN^[̊g嗦擾
VALUE SpriteCharacter_get_zoom ( VALUE self )
{
    BK::SpriteCharacter * const ptr = BK::SpriteCharacter::value_to_data( self );
    
    return rb_uint_new( ptr->get_zoom() );
}

/// S_̐ݒ
VALUE SpriteCharacter_set_center ( VALUE self, VALUE pointobj )
{
    using namespace BK;
    SpriteCharacter * const ptr = SpriteCharacter::value_to_data( self );
    const Point2 point = value_to_Point2( pointobj );
    
    try {
        ptr->set_center( point );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return pointobj;
}

/// g嗦̐ݒ
VALUE SpriteCharacter_set_zoom ( VALUE self, VALUE numobj )
{
    using namespace BK;
    SpriteCharacter * const ptr = SpriteCharacter::value_to_data( self );
    const int num = NUM2INT( numobj );
    
    try {
        ptr->set_zoom( num );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return numobj;
}

/// XvCg̐ݒ
VALUE SpriteCharacter_set_sprite ( VALUE self, VALUE nameobj )
{
    using namespace BK;
    SpriteCharacter * const ptr = SpriteCharacter::value_to_data( self );
    const char * const name = StringValuePtr( nameobj );
    
    try {
        ptr->set_sprite( name );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return nameobj;
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  SpriteCharacter
//

VALUE BK::SpriteCharacter::ruby_class_object = Qnil;

BK::SpriteCharacter::SpriteCharacter ( BK::SpriteField & field )
        : obj_field       (field),
          ptr_sprite      (NULL),
          ptr_projection  (NULL),
          point_center    (0, 0),
          num_zoom        (100),
          num_angle       (0),
          ptr_color       (NULL)
{
    BitmapProjectionLayer & layer = field.get_layer();
    
    this->ptr_projection = layer.register_projection();
    if( this->ptr_projection == NULL ) {
        throw new runtime_exception(
            "XvCgLN^ւ̓]̈抄蓖ĂɎs. (gpC[: %s)", typeid(layer).name());
    }
}

BK::SpriteCharacter::~SpriteCharacter ()
{
    if( !this->is_ruby_wrapped() ) {
        ProjectionInfo * const proj = this->ptr_projection;
        
        if( proj != NULL ) {
            proj->terminate();
            delete proj;
        }
    }
}

VALUE BK::SpriteCharacter::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "SpriteCharacter", parent_t::get_ruby_class() );
    
    xrb_define_0arg_method( c, "zoom", SpriteCharacter_get_zoom );
    xrb_define_1arg_method( c, "zoom=", SpriteCharacter_set_zoom );
    xrb_define_0arg_method( c, "center", SpriteCharacter_get_center );
    xrb_define_1arg_method( c, "center=", SpriteCharacter_set_center );
    xrb_define_1arg_method( c, "sprite=", SpriteCharacter_set_sprite );
    
    return ruby_class_object;
}

void BK::SpriteCharacter::ruby_gc_mark ( void )
{
    parent_t::ruby_gc_mark();
    
    ProjectionInfo * const proj = this->ptr_projection;
    
    if( proj != NULL ) {
        proj->gc_mark_self();
    }
}

void BK::SpriteCharacter::on_ruby_wrap ( void )
{
    parent_t::on_ruby_wrap();
    
    ProjectionInfo * const proj = this->ptr_projection;
    
    if( proj != NULL ) {
        proj->ruby_wrap();
    }
}

void BK::SpriteCharacter::on_activate ( BK::SpriteCharacter::stat_t status )
{
    parent_t::on_activate( status );
    
    ProjectionInfo * const proj = this->ptr_projection;
    
    if( proj != NULL ) {
        switch( status ) {
        case STAT_ACTIVE:
            proj->activate();
            break;
            
        case STAT_INACTIVE:
            proj->deactivate();
            break;
            
        case STAT_TERMINATED:
            proj->terminate();
            if( !this->is_ruby_wrapped() ) {
                delete proj;
            }
            this->ptr_projection = NULL;
            break;
        }
    }
}

void BK::SpriteCharacter::execute ( void )
{
    // ܂̓XNvgsĂ
    parent_t::execute();
    
    // Wɕϊ
    const SpriteFieldData * const sprite = this->ptr_sprite;
    ProjectionInfo * const proj = this->ptr_projection;
    
    if( proj != NULL ) {
        if( proj->is_active() ) {
            // TODO: zoom T|[g
            // TODO: angle T|[g
            if( sprite != NULL ) {
                proj->set_source( *sprite );
                proj->set_destination( Rect2(sprite->get_size()) + (this->point_center - sprite->get_center()) );
            }
            proj->set_color( this->ptr_color );
        }
    }
}

void BK::SpriteCharacter::set_sprite ( const char * name )
{
    if( name == NULL ) {
        throw new argument_exception("XvCgLN^gpXvCg NULL w肳܂.");
    }
    
    SpriteField & field = this->obj_field;
    const SpriteFieldData * const sprite = field.find( name );
    
    if( sprite == NULL ) {
        throw new argument_exception(
            "XvCgtB[h \"%s\" ɃXvCg \"%s\" o^.", field.get_name(), name);
    }
    this->ptr_sprite = sprite;
}

void BK::SpriteCharacter::set_center ( const BK::Point2 & p )
{
    this->point_center = p;
}

void BK::SpriteCharacter::set_color ( const BK::AlphaColor * color )
{
    if( color == NULL ) {
        if( this->ptr_color != NULL ) {
            delete this->ptr_color;
            this->ptr_color = NULL;
        }
    }
    else {
        if( this->ptr_color == NULL ) {
            this->ptr_color = new AlphaColor (*color);
        } else {
            *(this->ptr_color) = *color;
        }
    }
}

void BK::SpriteCharacter::set_zoom ( unsigned int n )
{
    this->num_zoom = n;
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_application_character_SpriteCharacter ( void )
{
    SpriteCharacter::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_application_character_SpriteCharacter ( void )
{
}

//////////////////////////////////////////////////////////////////////
