#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  layoutvalue_t
//

std::string BK::layoutvalue_t::to_string ( void ) const
{
    char buf [ 128 ];
    
    if( is_pixel() ) {
        sprintf( buf, "%d", get_pixel() );
        return std::string(buf);
    }
    if( is_percent() ) {
        if( get_percent() == static_cast< float >( static_cast< int >(get_percent()) ) ) {
            sprintf( buf, "%d%%", static_cast< int >( get_percent() ) );
        } else {
            sprintf( buf, "%.1f%%", get_percent() );
        }
        return std::string(buf);
    }
    return "";
}

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::Layout
//
namespace {

/// CAEg̃TCY擾
VALUE Layout_get_size ( VALUE self )
{
    using namespace BK;
    Layout * const ptr = Layout::value_to_data( self );
    
    return ptr->get_size().create_ruby_object();
}

/// CAEg̈ʒu擾
VALUE Layout_get_point ( VALUE self )
{
    using namespace BK;
    Layout * const ptr = Layout::value_to_data( self );
    
    return ptr->get_point().create_ruby_object();
}

/// CAEg̋`擾
VALUE Layout_get_area ( VALUE self )
{
    using namespace BK;
    Layout * const ptr = Layout::value_to_data( self );
    
    return Rect2(ptr->get_point(), ptr->get_size()).create_ruby_object();
}

/// CAEg̈ʒuݒ肷
VALUE Layout_set_point ( int argc, VALUE * argv, VALUE self )
{
    using namespace BK;
    Layout * const ptr = Layout::value_to_data( self );
    Point2 point;
    
    try {
        Ruby::ArgScanner scanner("Layout#point=", argc, argv);
        
        point = scanner.fetch_Point2("CAEgʒu");
        
        scanner.finish();
        
        ptr->set_point( point );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return point.create_ruby_object();
}

/// CAEg̍XV
VALUE Layout_update ( VALUE self )
{
    BK::Layout * const ptr = BK::Layout::value_to_data( self );
    
    return (ptr->update()) ? Qtrue : Qfalse;
}

/// CAEgXVĂ邩𒲂ׂ
VALUE Layout_is_modified ( VALUE self )
{
    BK::Layout * const ptr = BK::Layout::value_to_data( self );
    
    return (ptr->is_modified()) ? Qtrue : Qfalse;
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::Layout
//

VALUE BK::Layout::ruby_class_object = Qnil;

BK::Layout::Layout ( void )
        : flag_modified  (false),
          point_layout   (0, 0),
          size_layout    (0, 0)
{
}

BK::Layout::~Layout ()
{
}

VALUE BK::Layout::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "Layout", RubyObject::get_ruby_class() );
    
    xrb_define_0arg_method( c, "size", Layout_get_size );
    xrb_define_0arg_method( c, "point", Layout_get_point );
    xrb_define_0arg_method( c, "area", Layout_get_area );
    xrb_define_narg_method( c, "point=", Layout_set_point );
    xrb_define_0arg_method( c, "update", Layout_update );
    xrb_define_0arg_method( c, "modified?", Layout_is_modified );
    
    VALUE volatile value;
    
    value = INT2NUM( NONE );
    rb_define_const( c, "NONE", value );
    
    value = INT2NUM( LEFT );
    rb_define_const( c, "LEFT", value );
    
    value = INT2NUM( TOP );
    rb_define_const( c, "TOP", value );
    
    value = INT2NUM( CENTER );
    rb_define_const( c, "CENTER", value );
    
    value = INT2NUM( MIDDLE );
    rb_define_const( c, "MIDDLE", value );
    
    value = INT2NUM( RIGHT );
    rb_define_const( c, "RIGHT", value );
    
    value = INT2NUM( BOTTOM );
    rb_define_const( c, "BOTTOM", value );
    
    value = INT2NUM( FIT );
    rb_define_const( c, "FIT", value );
    
    return ruby_class_object;
}

const char * BK::Layout::align_to_string ( BK::Layout::align_t align )
{
    switch( align )
    {
    case NONE:
        return "NONE";
        
    case LEFT:
        return "LEFT";
        
    case CENTER:
        return "CENTER";
        
    case RIGHT:
        return "RIGHT";
        
    case TOP:
        return "TOP";
        
    case MIDDLE:
        return "MIDDLE";
        
    case BOTTOM:
        return "BOTTOM";
        
    case FIT:
        return "FIT";
        
    default:
        return "UNKNOWN";
    }
}

BK::Layout::align_t BK::Layout::string_to_align ( const char * str )
{
    if( str == NULL ) {
        return NONE;
    }
    if( strcmp(str, "left") == 0 ) {
        return LEFT;
    }
    if( strcmp(str, "center") == 0 ) {
        return CENTER;
    }
    if( strcmp(str, "right") == 0 ) {
        return RIGHT;
    }
    if( strcmp(str, "top") == 0 ) {
        return TOP;
    }
    if( strcmp(str, "middle") == 0 ) {
        return MIDDLE;
    }
    if( strcmp(str, "bottom") == 0 ) {
        return BOTTOM;
    }
    if( strcmp(str, "fit") == 0 ) {
        return FIT;
    }
    return NONE;
}

void BK::Layout::set_point ( const Point2 & p )
{
    #if 0
    if( log_is_trace() ) {
        log_trace("%s <%p>: %s  %s", typeid(*this).name(), this, point_layout.to_string().c_str(), p.to_string().c_str());
    }
    #endif
    if( point_layout != p ) {
        const Point2 delta = p - point_layout;
        point_layout = p;
        on_set_point( delta );
        set_modified();
    }
}

void BK::Layout::set_size ( const Size2 & s )
{
    if( size_layout != s ) {
        size_layout = s;
        set_modified();
    }
}

bool BK::Layout::update ( void )
{
    on_invoke_update();
    
    if( is_modified() ) {
        on_update();
        flag_modified = false;
        return true;
    } else {
        return false;
    }
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_utility_Layout ( void )
{
    BK::Layout::get_ruby_class();
    
    Init_utility_CharLayout();
    Init_utility_DocumentLayout();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_utility_Layout ( void )
{
    Exit_utility_DocumentLayout();
    Exit_utility_CharLayout();
}

//////////////////////////////////////////////////////////////////////
