#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  t@C[Jȃ\bh
//
namespace {

using namespace BK;

inline get_label_height ( const TextLayout & owner )
{
    const layoutvalue_t & interval = owner.get_line_interval();
    
    if( interval.is_null() ) {
        return owner.get_minimum_height();
    } else {
        return interval.get( owner.get_size().get_height() );
    }
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  t@C[JȃNX
//
namespace {

class LabelLayoutArray : public std::vector < LabelLayout * >
{
public:
    
    /// fXgN^
    ~LabelLayoutArray ()
    {
        iterator e = this->end();
        iterator itr;
        
        for( itr = this->begin(); itr != e; ++itr ) {
            delete *itr;
        }
    }
    
};

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::TextLayout
//
namespace {

/// IuWFNg̐
VALUE TextLayout_new ( int argc, VALUE * argv, VALUE klass )
{
    using namespace BK;
    VALUE volatile obj = Qnil;
    
    try {
        Ruby::ArgScanner scanner("TextLayout.new", argc, argv);
        Size2 size;
        
        size = scanner.fetch_Size2("CAEgTCY");
        
        scanner.finish();
        
        TextLayout * const ptr = new TextLayout ();
        
        ptr->set_size( size );
        
        obj = ptr->ruby_wrap();
        rb_obj_call_init( obj, argc, argv );
        
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return obj;
}

/// IuWFNg̏
VALUE TextLayout_initialize ( int argc, VALUE * argv, VALUE self )
{
    return self;
}

/// sԊu̎擾
VALUE TextLayout_get_line_interval ( VALUE self )
{
    using namespace BK;
    TextLayout * const ptr = TextLayout::value_to_data( self );
    
    const layoutvalue_t value = ptr->get_line_interval();
    
    char buf [ 32 ];
    int len;
    
    if( value.is_percent() ) {
        len = sprintf( buf, "%.2f%%", value.get_percent() );
        return rb_str_new(buf, len);
    }
    
    if( value.is_pixel() ) {
        len = sprintf( buf, "%d", value.get_pixel() );
        return rb_str_new(buf, len);
    }
    
    return Qnil;
}

/// sԊu̐ݒ
VALUE TextLayout_set_line_interval ( VALUE self, VALUE argobj )
{
    using namespace BK;
    TextLayout * const ptr = TextLayout::value_to_data( self );
    layoutvalue_t value;
    
    if( !NIL_P(argobj) ) {
        if( TYPE(argobj) == T_STRING ) {
            const char * const str = RSTRING(argobj)->ptr;
            const int len = RSTRING(argobj)->len;
            char * ptr;
            double n = strtod( str, &ptr );
            
            if( *ptr == '\0' ) {
                value.set_pixel( static_cast< int >( n ) );
            } else if ( *ptr == '%' ) {
                value.set_percent( static_cast< float >( n ) );
            } else {
                rb_raise(rb_eArgError, "߂łȂl \"%s\" w肳܂.", str);
            }
        } else {
            value.set_pixel( NUM2INT(argobj) );
        }
    }
    ptr->set_line_interval( value );
    
    return argobj;
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::TextLayout
//

VALUE BK::TextLayout::ruby_class_object = Qnil;

BK::TextLayout::TextLayout ( void )
{
}

BK::TextLayout::~TextLayout ()
{
}

VALUE BK::TextLayout::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "TextLayout", parent_t::get_ruby_class() );
    
    xrb_define_narg_singleton_method( c, "new", TextLayout_new );
    
    xrb_define_narg_method( c, "initialize", TextLayout_initialize );
    xrb_define_0arg_method( c, "line_interval", TextLayout_get_line_interval );
    xrb_define_1arg_method( c, "line_interval=", TextLayout_set_line_interval );
    
    return ruby_class_object;
}

void BK::TextLayout::on_update ( void )
{
    parent_t::on_update();
    
    // ZbgAbv
    if( ary_logline.empty() ) {
        return;
    }
    
    Point2 point = get_point();
    const Size2 size = get_size();
    const Rect2 layoutarea ( point, size );
    
    if( !is_overflow_enabled() ) {
        if( !layoutarea.is_include(point) ) {
            return;
        }
    }
    unsigned int count = count_document_layouts();
    auto_ary< Layout * > layoutsbuf ( count ); // 郊XgłȂƃ[N
    Layout * * const layouts = layoutsbuf.get();
    
    count = get_document_layouts( layouts, count );
    
    // xXg̐
    #if 0
    log_trace("xXg̐Jn܂.");
    #endif
    const Size2 labelsize ( size.get_width(), get_label_height(*this) );
    LabelLayoutArray labelary;
    {
        Layout * * layoutptr = layouts;
        lineary_t lineary ( ary_logline ); // _sXg̃Rs[g
        unsigned int total = 0;
        
        while( !lineary.empty() ) {
            std::auto_ptr < LabelLayout > label ( new LabelLayout() );
            label->set_size( labelsize );
            label->set_horizontal_align( get_horizontal_align() );
            label->set_vertical_align( Layout::BOTTOM );
            label->set_point( point );
            label->add_layout( layoutptr, lineary.front() );
            label->update();
            
            const unsigned int registered = label->count_registered_layouts();
            
            if( lineary.front() != 0 && registered == 0 ) {
                throw new runtime_exception("eLXgCAEg̉sĂ܂.");
            }
            point.add_y( labelsize.get_height() );
            if( !is_overflow_enabled() ) {
                if( !layoutarea.is_include(point) ) {
                    break;
                }
            }
            labelary.push_back( label.release() );
            
            lineary.front() -= registered;
            layoutptr += registered;
            total += registered;
            if( lineary.front() == 0 ) {
                lineary.erase( lineary.begin() );
            }
        }
        set_registered_layouts_count( total );
    }
    
    // xXg̈ʒu
    #if 0
    log_trace("xXg̈ʒuC܂.");
    #endif
    {
        const LabelLayoutArray::iterator e = labelary.end();
        LabelLayoutArray::iterator itr;
        Layout::align_t align = get_vertical_align();
        Layout * layout;
        
        if( align == Layout::FIT ) {
            if( labelary.size() == 1 ) {
                align = Layout::CENTER;
            }
        }
        
        switch( align ) {
        case Layout::FIT:
            {
                const Point2::num_t basetop = get_point().get_y();
                const int h = size.get_height() - labelary.back()->get_size().get_height();
                const unsigned int allcount = labelary.size() - 1;
                unsigned int count = 0;
                
                for( itr = labelary.begin(); itr != e; ++itr ) {
                    layout = *itr;
                    point = layout->get_point();
                    point.set_y( basetop + (h * count / allcount) );
                    ++count;
                    layout->set_point( point );
                }
            }
            break;
            
        case Layout::CENTER:
        case Layout::BOTTOM:
            {
                Size2::num_t offset;
                Size2::num_t contentheight = labelsize.get_height() * labelary.size();
                
                if( align == Layout::CENTER ) {
                    offset = (size.get_height() - contentheight) / 2;
                } else {
                    offset = size.get_height() - contentheight;
                }
                for( itr = labelary.begin(); itr != e; ++itr ) {
                    layout = *itr;
                    point = layout->get_point();
                    point.add_y( offset );
                    layout->set_point( point );
                }
            }
            break;
            
        case Layout::TOP:
        default:
            // ㍇킹Ȃ̂ŉȂ
            break;
        }
    }
}

void BK::TextLayout::set_line_interval ( const BK::layoutvalue_t & val )
{
    obj_line_interval = val;
}

void BK::TextLayout::add_layout ( BK::Layout * * layouts, unsigned int len )
{
    DocumentLayout::add_layout( layouts, len );
    ary_logline.push_back( len );
}

void BK::TextLayout::clear_registered_layouts ( void )
{
    unsigned int registered = count_registered_layouts();
    
    DocumentLayout::clear_registered_layouts();
    
    while( registered > 0 ) {
        if( registered >= ary_logline.front() ) {
            registered -= ary_logline.front();
            ary_logline.erase( ary_logline.begin() );
        } else {
            ary_logline.front() -= registered;
            registered = 0;
        }
    }
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_utility_TextLayout ( void )
{
    BK::TextLayout::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_utility_TextLayout ( void )
{
}

//////////////////////////////////////////////////////////////////////
