#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  O[o֐
//
namespace {

const int SPRINTF_BUFFER_CAPACITY = 16 * 1024;

} /* namespace */

std::string BK::strformat ( const char * format, ... )
{
    va_list arg;
    char buffer [SPRINTF_BUFFER_CAPACITY];
    
    va_start( arg, format );
    vsprintf(buffer, format, arg);
    va_end( arg );
    
    return std::string(buffer);
}

std::string BK::strvformat ( const char * format, va_list arg )
{
    char buffer [SPRINTF_BUFFER_CAPACITY];
    
    vsprintf(buffer, format, arg);
    
    return std::string(buffer);
}

unsigned int BK::str_translate_char ( char * str, char from, char to )
{
    unsigned int ret = 0;
    char c;
    
    switch( rb_get_kcode()[0] ) {
    case 'S': case 's':
        {
            while( (c = *str) != '\0' ) {
                if( is_sjis_lead_char(c) ) {
                    str += 2;
                } else {
                    if( c == from ) {
                        *str = to;
                        ++ret;
                    }
                    str += 1;
                }
            }
        }
        break;
        
    case 'E': case 'e':
        {
            while( (c = *str) != '\0' ) {
                if( is_euc_lead_char(c) ) {
                    str += 2;
                } else {
                    if( c == from ) {
                        *str = to;
                        ++ret;
                    }
                    str += 1;
                }
            }
        }
        break;
    }
    
    return ret;
}

int BK::str_index_char ( const char * str, char ch )
{
    const char * const head = str;
    char c;
    
    switch( rb_get_kcode()[0] ) {
    case 'S': case 's':
        {
            while( (c = *str) != '\0' ) {
                if( is_sjis_lead_char(c) ) {
                    str += 2;
                } else {
                    if( c == ch ) {
                        return str - head;
                    }
                    str += 1;
                }
            }
        }
        break;
        
    case 'E': case 'e':
        {
            while( (c = *str) != '\0' ) {
                if( is_euc_lead_char(c) ) {
                    str += 2;
                } else {
                    if( c == ch ) {
                        return str - head;
                    }
                    str += 1;
                }
            }
        }
        break;
    }
    
    return -1;
}

BK::Size2 BK::value_to_Size2 ( VALUE obj )
{
    Ruby::ArgScanner scanner("Size2", 1, &obj);
    
    return scanner.fetch_Size2();
}

BK::Point2 BK::value_to_Point2 ( VALUE obj )
{
    Ruby::ArgScanner scanner("Point2", 1, &obj);
    
    return scanner.fetch_Point2();
}

BK::Rect2 BK::value_to_Rect2 ( VALUE obj )
{
    Ruby::ArgScanner scanner("Rect2", 1, &obj);
    
    return scanner.fetch_Rect2();
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_utility ( void )
{
    Init_utility_StringBuffer();
    Init_utility_OnOffObject();
    Init_utility_CriticalSection();
    Init_utility_Event();
    Init_utility_ServiceDevice();
    
    Init_utility_Size2();
    Init_utility_Point2();
    Init_utility_Rect2();
    Init_utility_AspectRect2();
    Init_utility_ProjectionInfo();
    Init_utility_Layout();
    
    Init_utility_ArgScanner();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_utility ( void )
{
    Exit_utility_ArgScanner();
    
    Exit_utility_Layout();
    Exit_utility_ProjectionInfo();
    Exit_utility_AspectRect2();
    Exit_utility_Rect2();
    Exit_utility_Point2();
    Exit_utility_Size2();
    
    Exit_utility_ServiceDevice();
    Exit_utility_Event();
    Exit_utility_CriticalSection();
    Exit_utility_OnOffObject();
    Exit_utility_StringBuffer();
}

//////////////////////////////////////////////////////////////////////
