#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  t@C[JȒ萔
//
namespace {

const unsigned int RUBY_INSPECT_BUFFER_SIZE = 4096;

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: RubyObject
//
namespace {

using namespace BK;

void RubyObject_mark ( RubyObject * ptr )
{
    #if 0
    VALUE const obj = ptr->get_ruby_object();
    if( (RBASIC(obj)->flags & FL_MARK) == 0 ) {
        log_error("}[ÑIuWFNgɑ΂ RubyObject_mark(%08u) Ăяoꂽ.", obj);
    }
    #endif
    #if 0
    if( log_is_trace() ) {
        log_trace("Ruby GC  %s <%p> (%08u) }[N܂.", typeid(*ptr).name(), ptr, ptr->get_ruby_object());
        ptr->ruby_gc_mark();
        return;
    }
    #endif
    ptr->ruby_gc_mark();
}

void RubyObject_free ( RubyObject * ptr )
{
    if( log_is_debug() && ptr->is_ruby_gc_log_enabled() ) {
        log_debug("Ruby GC  %s <%p> ܂.", typeid(*ptr).name(), ptr);
    }
    delete ptr;
}

VALUE RubyObject_inspect ( VALUE self )
{
    using namespace BK;
    const unsigned int size = RUBY_INSPECT_BUFFER_SIZE;
    char buf [ size ];
    
    RubyObject * const ptr = RubyObject::value_to_data( self );
    
    char * result = ptr->ruby_inspect( buf, size );
    if( result == NULL ) {
        return rb_call_super( 0, NULL );
    }
    
    VALUE volatile obj = rb_str_new2( result );
    if( result < buf || buf + size <= result ) {
        delete result;
    }
    
    return obj;
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::RubyObject
//

VALUE BK::RubyObject::ruby_class_object = Qnil;

VALUE BK::RubyObject::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "RubyObject", rb_cData );
    
    xrb_define_0arg_method( c, "inspect", RubyObject_inspect );
    
    return ruby_class_object;
}

void * BK::RubyObject::get_ruby_mark_function ( void )
{
    return RubyObject_mark;
}

void * BK::RubyObject::get_ruby_free_function ( void )
{
    return RubyObject_free;
}

void BK::RubyObject::set_ruby_object ( VALUE obj )
{
    if( !NIL_P(obj_ruby_body) ) {
        throw new logic_exception("%s ɑ΂ Ruby IuWFNg̍Đݒ肪sꂽ.", typeid(*this).name());
    }
    obj_ruby_body = obj;
}

char * BK::RubyObject::ruby_inspect ( char * buf, unsigned int cap ) const
{
    VALUE const v = this->obj_ruby_body;
    
    if( NIL_P(v) ) {
        sprintf(buf, "#<%s:%p>", typeid(*this).name(), this);
    } else {
        sprintf(buf, "#<%s:0x%x|%s:%p>", rb_obj_classname(v), v, typeid(*this).name(), this);
    }
    return buf;
}

void BK::RubyObject::gc_mark_self ( void )
{
    if( is_ruby_wrapped() ) {
        rb_gc_mark( this->obj_ruby_body );
    } else {
        this->ruby_gc_mark();
    }
}

VALUE BK::RubyObject::ruby_wrap ( void )
{
    if( !NIL_P(this->get_ruby_object()) ) {
        return this->get_ruby_object();
    }
    
    VALUE  const klass     = this->get_ruby_class_object();
    void * const markfunc  = this->get_ruby_mark_function();
    void * const freefunc  = this->get_ruby_free_function();
    void * const thisptr   = this->get_wrapped_address();
    
    VALUE volatile obj = Data_Wrap_Struct( klass, markfunc, freefunc, thisptr );
    this->set_ruby_object( obj );
    
    this->on_ruby_wrap();
    
    return obj;
}

std::string BK::RubyObject::get_ruby_inspect_string ( void ) const
{
    const unsigned int size = RUBY_INSPECT_BUFFER_SIZE;
    char buf [ size ];
    
    char * result = ruby_inspect( buf, size );
    if( result == NULL ) {
        return std::string();
    }
    
    std::string ret ( buf );
    
    if( result < buf || buf + size <= result ) {
        delete result;
    }
    
    return ret;
}

//////////////////////////////////////////////////////////////////////
////  BK::ruby_exception
//

BK::ruby_exception::ruby_exception ( VALUE obj, const char * format, ... )
        : obj_error (Qnil)
{
    va_list arg;
    
    va_start( arg, format );
    initialize( obj, format, arg );
    va_end( arg );
}

BK::ruby_exception::ruby_exception ( const char * format, ... )
        : obj_error (Qnil)
{
    va_list arg;
    
    va_start( arg, format );
    initialize( ruby_errinfo, format, arg );
    va_end( arg );
}

BK::ruby_exception::~ruby_exception ()
{
    if( !NIL_P(this->obj_error) ) {
        rb_gc_unregister_address( &(this->obj_error) );
    }
}

void BK::ruby_exception::initialize ( VALUE obj, const char * format, va_list arg )
{
    obj_error = obj;
    
    // unregister ̓fXgN^ōsĂ܂
    rb_gc_register_address( &obj_error );
    
    this->set_string( format, arg );
    
    const char * mes = StringValuePtr( obj );
    
    this->the_string().append( BK::strformat(" [RubyO: %s]", mes) );
}

//////////////////////////////////////////////////////////////////////
////  O[o֐ (Ruby f[^ϊ)
//

bool xrb_is_kind_of_any ( VALUE obj, ... )
{
    va_list arg;
    VALUE klass;
    bool result = false;
    
    va_start( arg, obj );
    while( !result ) {
        klass = va_arg(arg, VALUE);
        if( NIL_P(klass) ) {
            break;
        }
        result = xrb_is_kind_of( obj, klass );
    }
    va_end( arg );
    
    return result;
}

//////////////////////////////////////////////////////////////////////
////  O[o֐ (BK-Ruby Ag)
//
namespace {

/// C++ f[^i[Ă߂̃z_IuWFNg
struct CppData
{
    union
    {
        int            value_int;
        unsigned int   value_uint;
        long           value_long;
        unsigned long  value_ulong;
        double         value_double;
    };
    VALUE  source_data;
};

void _translation_fail ( VALUE val, const char * type )
{
    VALUE volatile errinfo = ruby_errinfo;
    VALUE errobj = errinfo;
    const char * errmes;
    
    errmes = StringValuePtr(errobj);
    throw new BK::argument_exception(
        "%s  %s ւ̕ϊɎs. ([%s] %s)",
        rb_obj_classname(val), type, rb_obj_classname(errinfo), errmes);
}

VALUE _value_to_int ( VALUE datobj )
{
    CppData * const ptr = ruby_cast< CppData >( datobj );
    
    VALUE volatile numobj = rb_Integer( ptr->source_data );
    
    ptr->value_int = NUM2INT( numobj );
    
    return datobj;
}

VALUE _value_to_uint ( VALUE datobj )
{
    CppData * const ptr = ruby_cast< CppData >( datobj );
    
    VALUE volatile numobj = rb_Integer( ptr->source_data );
    
    ptr->value_uint = NUM2UINT( numobj );
    
    return datobj;
}

VALUE _value_to_double ( VALUE datobj )
{
    CppData * const ptr = ruby_cast< CppData >( datobj );
    
    VALUE volatile numobj = rb_Float( ptr->source_data );
    
    ptr->value_double = NUM2DBL( numobj );
    
    return datobj;
}

} /* namespace */

void BK::ruby_throw ( BK::exception_base * e )
{
    const int typebuffer_cap = 256 - 1;
    const int mesbuffer_cap = 1024 * 4 - 1;
    char typebuffer [ typebuffer_cap + 1 ];
    char mesbuffer  [ mesbuffer_cap + 1 ];
    VALUE rubyclass = rb_eRuntimeError;
    
    log_trace("BK::ruby_throw(): [%s] %s", e->get_type(), e->what());
    {
        type_info const & type = typeid(*e);
        
        if( type == typeid(system_exit) ) {
            rubyclass = rb_eSystemExit;
        }
        else if( type == typeid(index_exception) ) {
            rubyclass = rb_eIndexError;
        }
        else if( type == typeid(argument_exception) ) {
            rubyclass = rb_eArgError;
        }
        else if( type == typeid(out_of_range_exception) ) {
            rubyclass = rb_eRangeError;
        }
        else if( type == typeid(no_memory_exception) ) {
            rubyclass = rb_eNoMemError;
        }
        else if( type == typeid(unsupported_exception) ) {
            rubyclass = rb_eNotImpError;
        }
        else if( type == typeid(runtime_exception) ) {
            rubyclass = rb_eRuntimeError;
        }
        else if( type == typeid(standard_exception) ) {
            rubyclass = rb_eStandardError;
        }
        
        strncpy( typebuffer, e->get_type(), typebuffer_cap );
        typebuffer[ typebuffer_cap ] = '\0';
        
        strncpy( mesbuffer, e->what(), mesbuffer_cap );
        mesbuffer[ mesbuffer_cap ] = '\0';
    }
    delete e;
    rb_raise(rubyclass, "[%s] %s", typebuffer, mesbuffer );
}

VALUE BK::get_ruby_bk_module ( void )
{
    return rb_define_module("BK");
}

VALUE BK::define_ruby_bk_class ( const char * name, VALUE superclass )
{
    const VALUE mod_bk = get_ruby_bk_module();
    
    return rb_define_class_under( mod_bk, name, superclass );
}

int BK::value_to_int ( VALUE val )
{
    if( FIXNUM_P(val) ) {
        return FIX2INT(val);
    }
    
    CppData dat;
    
    dat.source_data = val;
    
    VALUE volatile datobj = Data_Wrap_Struct( rb_cData, NULL, NULL, &dat );
    
    int state;
    
    rb_protect( _value_to_int, datobj, &state );
    if( state ) {
        _translation_fail( val, "int" );
    }
    
    return dat.value_int;
}

unsigned int BK::value_to_uint ( VALUE val )
{
    if( FIXNUM_P(val) ) {
        return FIX2UINT(val);
    }
    
    CppData dat;
    
    dat.source_data = val;
    
    VALUE volatile datobj = Data_Wrap_Struct( rb_cData, NULL, NULL, &dat );
    
    int state;
    
    rb_protect( _value_to_uint, datobj, &state );
    if( state ) {
        _translation_fail( val, "unsigned int" );
    }
    
    return dat.value_uint;
}

double BK::value_to_double ( VALUE val )
{
    if( TYPE(val) == T_FLOAT ) {
        return NUM2DBL(val);
    }
    
    CppData dat;
    
    dat.source_data = val;
    
    VALUE volatile datobj = Data_Wrap_Struct( rb_cData, NULL, NULL, &dat );
    
    int state;
    
    rb_protect( _value_to_double, datobj, &state );
    if( state ) {
        _translation_fail( val, "double" );
    }
    
    return dat.value_double;
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_ruby ( void )
{
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_ruby ( void )
{
}

//////////////////////////////////////////////////////////////////////
