#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  t@C[Jȃ\bh
//
namespace {

void copy_film_image ( BK::Film::pixel_t * dest, unsigned int wpitch, BK::Film::pixel_t const * src, BK::Size2 const & size )
{
    typedef BK::Film::pixel_t pixel_t;
    
    const unsigned int rpitch = size.get_width();
    unsigned int y;
    
    pixel_t * wlineptr = dest;
    for( y = size.get_height(); y > 0; --y ) {
        memcpy( wlineptr, src, sizeof(pixel_t) * size.get_width() );
        src += rpitch;
        wlineptr += wpitch;
    }
}

} /* namespace */
//////////////////////////////////////////////////////////////////////
////  Ruby NX: BK::Film
//
namespace {
extern "C" {

/// tBIuWFNg̍\z
VALUE Film_new ( int argc, VALUE * argv, VALUE klass )
{
    using namespace BK;
    VALUE volatile obj = Qnil;
    Size2 size;
    VALUE pattern;
    
    try {
        Ruby::ArgScanner scanner("Film.new", argc, argv);
        
        size = scanner.fetch_Size2("tBTCY");
        pattern = scanner.fetch_value("tBf[^");
        
        scanner.finish();
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    pattern = rb_check_array_type( pattern );
    if( NIL_P(pattern) ) {
        rb_raise(rb_eArgError, "tBf[^ Array ȊO (%s) ͎wł܂.", rb_obj_classname(pattern));
    }
    
    const unsigned int area = size.get_area();
    const unsigned int len = static_cast< unsigned int >( xrb_ary_length(pattern) );
    
    if( area != len ) {
        rb_raise(rb_eArgError, "f[^̃TCY (%u) ƃtB̃TCY (%u) v܂.", len, area);
    }
    Film::pixel_t * const img = ALLOCA_N(Film::pixel_t, area);
    
    for( unsigned int i = 0; i < len; ++i ) {
        const VALUE v = xrb_ary_at(pattern, i);
        
        if( !FIXNUM_P(v) ) {
            rb_raise(rb_eArgError, "tBf[^ Fixnum ȊO (%s) ͎wł܂.", rb_obj_classname(v));
        }
        
        const unsigned int pixelval = FIX2UINT(v);
        
        if( pixelval > BK::COLOR_ELEMENT_MAX ) {
            rb_raise(rb_eArgError, "tBf[^Ɏw肳ꂽl (%u) ͈͊Oł.", pixelval);
        }
        img[i] = static_cast< Film::pixel_t >( pixelval );
    }
    
    try {
        Film * const film = new Film( size, img );
        obj = film->ruby_wrap();
        rb_obj_call_init( obj, argc, argv );
    }
    catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return obj;
}

/// tBIuWFNg̏
VALUE Film_initialize ( int argc, VALUE * argv, VALUE self )
{
    return self;
}

/// tB̃TCY擾
VALUE Film_get_size ( VALUE self )
{
    BK::Film * const film = BK::Film::value_to_data( self );
    
    return film->get_size().create_ruby_object();
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::Film
//

VALUE BK::Film::ruby_class_object = Qnil;

BK::Film::Film ( const BK::Size2 & size )
        : size_image  (size),
          ary_image   (new pixel_t [ size.get_area() ])
{
    memset( ary_image, 0, sizeof(pixel_t) * size.get_area() );
}

BK::Film::Film ( const BK::Size2 & size, const pixel_t * img )
        : size_image  (size),
          ary_image   (new pixel_t [ size.get_area() ])
{
    memcpy( ary_image, img, sizeof(pixel_t) * size.get_area() );
}

BK::Film::~Film ()
{
    delete[] ary_image;
}

VALUE BK::Film::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "Film", parent_t::get_ruby_class() );
    
    xrb_define_narg_singleton_method( c, "new", Film_new );
    
    xrb_define_narg_method( c, "initialize", Film_initialize );
    xrb_define_0arg_method( c, "size", Film_get_size );
    
    return ruby_class_object;
}

void BK::Film::set_size ( const BK::Size2 & s )
{
    pixel_t * const ptr = new pixel_t [ s.get_area() ];
    pixel_t * rptr = ary_image;
    pixel_t * wptr = ptr;
    const Size2::num_t h = std::min( s.get_height(), size_image.get_height() );
    const Size2::num_t w = std::min( s.get_width(), size_image.get_width() );
    const Size2::num_t zh = s.get_height() - h;
    const Size2::num_t zw = s.get_width() - w;
    int y;
    
    for( y = 0; y < h; ++y ) {
        memcpy( &wptr[0], rptr, sizeof(pixel_t) * w );
        memset( &wptr[w], 0, sizeof(pixel_t) * zw );
        wptr += s.get_width();
        rptr += size_image.get_width();
    }
    for( y = 0; y < zh; ++y ) {
        memset( &wptr[0], 0, sizeof(pixel_t) * s.get_width() );
        wptr += s.get_width();
    }
    
    delete[] ary_image;
    size_image = s;
    ary_image = ptr;
}

void BK::Film::set_image ( BK::Film::pixel_t const * img, BK::Rect2 const & r )
{
    const unsigned int imgpitch = get_size().get_width();
    Rect2 rect = r;
    
    // sNZf[^Rs[
    //rect.normalize(); // \bh̋KŐKς
    rect.clipping( Rect2(get_size()) );
    
    copy_film_image(ary_image + at_image_offset(rect.get_point()), imgpitch, img, rect.get_size());
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_graphic_Film ( void )
{
    BK::Film::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_graphic_Film ( void )
{
}

//////////////////////////////////////////////////////////////////////
