#include "stdafx.h"
#include "schema/main.h"
#include <string.h>

//////////////////////////////////////////////////////////////////////
////  Ruby NX: BK::Glyph
//
namespace {
extern "C" {

/// ւ̕ϊ
VALUE Glyph_to_s ( VALUE self )
{
    BK::Glyph * const glyph = BK::Glyph::value_to_data( self );
    
    return rb_str_new2(glyph->get_character());
}

/// tB擾
VALUE Glyph_get_film ( VALUE self )
{
    BK::Glyph * const glyph = BK::Glyph::value_to_data( self );
    VALUE volatile obj = Qnil;
    
    try {
        obj = glyph->get_film()->ruby_wrap();
    }
    catch ( BK::exception_base * e ) {
        BK::ruby_throw( e );
    }
    
    return obj;
}

/// IuWFNg̃TCY擾
VALUE Glyph_get_size ( VALUE self )
{
    BK::Glyph * const glyph = BK::Glyph::value_to_data( self );
    BK::Size2 s;
    
    try {
        BK::Glyph::ABC_t const & abc = glyph->get_ABC();
        s = glyph->get_film()->get_size();
        s.set_width( ((abc.a < 0) ? 0 : abc.a) + abc.b + ((abc.c < 0) ? 0 : abc.c) );
    }
    catch ( BK::exception_base * e ) {
        BK::ruby_throw( e );
    }
    
    return s.create_ruby_object();
}

/// ̑OɕKvȋ󔒗ʂ擾
VALUE Glyph_get_front_offset ( VALUE self )
{
    BK::Glyph * const glyph = BK::Glyph::value_to_data( self );
    int offset;
    
    try {
        BK::Glyph::ABC_t const & abc = glyph->get_ABC();
        if( abc.a < 0 ) {
            offset = -(abc.a);
        } else {
            offset = 0;
        }
    }
    catch ( BK::exception_base * e ) {
        BK::ruby_throw( e );
    }
    
    return rb_fix_new( offset );
}

/// ̑OɕKvȋ󔒗ʂ擾
VALUE Glyph_get_back_offset ( VALUE self )
{
    BK::Glyph * const glyph = BK::Glyph::value_to_data( self );
    int offset;
    
    try {
        BK::Glyph::ABC_t const & abc = glyph->get_ABC();
        if( abc.c < 0 ) {
            offset = -(abc.c);
        } else {
            offset = 0;
        }
    }
    catch ( BK::exception_base * e ) {
        BK::ruby_throw( e );
    }
    
    return rb_fix_new( offset );
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::Glyph
//

VALUE BK::Glyph::ruby_class_object = Qnil;

BK::Glyph::Glyph ( const char * str, Film * img, const BK::Glyph::ABC_t & abc )
        : str_character  ((str) ? str : ""),
          obj_abc        (abc),
          ptr_image      (img)
{
    if( this->ptr_image == NULL ) {
        throw new argument_exception("OtIuWFNg̎gp摜 NULL w肳܂.");
    }
}

BK::Glyph::~Glyph ( void )
{
}

VALUE BK::Glyph::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "Glyph", BK::RubyObject::get_ruby_class() );
    
    // Glyph  new łȂ
    
    xrb_define_0arg_method( c, "to_s", Glyph_to_s );
    xrb_define_0arg_method( c, "film", Glyph_get_film );
    xrb_define_0arg_method( c, "size", Glyph_get_size );
    xrb_define_0arg_method( c, "front_offset", Glyph_get_front_offset );
    xrb_define_0arg_method( c, "back_offset", Glyph_get_back_offset );
    
    return ruby_class_object;
}

void BK::Glyph::ruby_gc_mark ( void )
{
    this->ptr_image->gc_mark_self();
}

char * BK::Glyph::ruby_inspect ( char * buf, unsigned int cap ) const
{
    sprintf(buf, "#<BK::Glyph:%p (char=%s,film=%p)>", this, str_character.c_str(), ptr_image);
    return buf;
}

void BK::Glyph::on_ruby_wrap ( void )
{
    this->ptr_image->ruby_wrap();
}

//////////////////////////////////////////////////////////////////////
////  Ruby NX: BK::Font
//
namespace {
extern "C" {

/// tHg̍쐬
VALUE Font_new ( int argc, VALUE * argv, VALUE klass )
{
    using namespace BK;
    VALUE volatile nameobj = Qnil;
    VALUE volatile sizeobj = Qnil;
    VALUE volatile paramobj = Qnil;
    Font::info_t info;
    
    try {
        Ruby::ArgScanner scanner("Font.new", argc, argv);
        
        nameobj = scanner.fetch_value("tHg");
        sizeobj = scanner.fetch_value("tHgTCY");
        if( scanner.is_empty() ) {
            paramobj = Qnil;
        } else {
            paramobj = scanner.fetch_value("tHgp[^");
        }
        scanner.finish();
    } catch ( exception_base * e ) {
        ruby_throw(e);
    }
    if( !NIL_P(paramobj) && TYPE(paramobj) != T_HASH ) {
        rb_raise(rb_eArgError, "tHg̃p[^e[u Hash ȊO (%s) ͎wł܂.",
                 rb_obj_classname(paramobj));
    }
    
    info.font_name    = StringValuePtr( nameobj );
    info.font_height  = NUM2INT(sizeobj);
    
    if( !NIL_P(paramobj) ) {
        {
            const VALUE valobj = rb_hash_aref(paramobj, rb_str_new2("Proportional"));
            
            if( RTEST(valobj) ) {
                info.is_proportional = true;
            }
        }
    }
    
    VALUE volatile obj = Qnil;
    
    try {
        Font * const font = create_Font( info );
        obj = font->ruby_wrap();
        rb_obj_call_init( obj, argc, argv );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return obj;
}

/// tHg̏
VALUE Font_initialize ( int argc, VALUE * argv, VALUE self )
{
    return self;
}

/// tHgLbṼTCY擾
VALUE Font_get_cache_size ( VALUE klass )
{
    BK::FontDatabase * const db = BK::FontDatabase::get_instance();
    
    return rb_uint_new(db->get_cache_size());
}

/// tHgLbṼTCYݒ肷
VALUE Font_set_cache_size ( VALUE klass, VALUE numobj )
{
    BK::FontDatabase * const db = BK::FontDatabase::get_instance();
    
    db->set_cache_size( NUM2UINT(numobj) );
    
    return numobj;
}

/// tHg擾
VALUE Font_get_name ( VALUE self )
{
    BK::Font * const font = BK::Font::value_to_data( self );
    
    return rb_str_new2(font->get_font_name());
}

/// OtIuWFNg̔z擾
VALUE Font_fetch ( VALUE self, VALUE strobj )
{
    using namespace BK;
    Font * const font = Font::value_to_data( self );
    const volatile VALUE ret = rb_ary_new();
    
    try {
        const char * str = StringValuePtr( strobj );
        
        while( *str != '\0' ) {
            #if 0
            if( log_is_trace() ) {
                log_trace("font->get_glyph(\"%s\")", str);
            }
            #endif
            Glyph * const glyph = font->get_glyph( str );
            
            if( glyph != NULL ) {
                rb_ary_push( ret, glyph->ruby_wrap() );
                str += strlen( glyph->get_character() );
            } else {
                if( is_sjis_lead_char(str[0]) || is_euc_lead_char(str[0]) ) {
                    str += 2;
                } else {
                    str += 1;
                }
            }
        }
    }
    catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return ret;
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::Font::info_t
//

BK::Font::info_t::info_t ( void )
        : font_height      (-1),
          is_proportional  (false),
          font_weight      (0),
          is_italic        (false),
          is_raster_prior  (false)
{
}

//////////////////////////////////////////////////////////////////////
////  BK::Font
//

VALUE BK::Font::ruby_class_object = Qnil;

VALUE BK::Font::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "Font", parent_t::get_ruby_class() );
    
    xrb_define_narg_singleton_method( c, "new", Font_new );
    xrb_define_0arg_singleton_method( c, "cache_size", Font_get_cache_size );
    xrb_define_1arg_singleton_method( c, "cache_size=", Font_set_cache_size );
    
    xrb_define_narg_method( c, "initialize", Font_initialize );
    xrb_define_0arg_method( c, "name", Font_get_name );
    xrb_define_1arg_method( c, "fetch", Font_fetch );
    xrb_define_1arg_method( c, "[]", Font_fetch );
    
    return ruby_class_object;
}

BK::Font::Font ( void )
{
}

BK::Font::~Font ()
{
}

BK::Glyph * BK::Font::get_glyph ( const char * str )
{
    FontDatabase * const db = FontDatabase::get_instance();
    
    if( db == NULL ) {
        return create_glyph( str );
    }
    
    return db->get_glyph( *this, str );
}

//////////////////////////////////////////////////////////////////////
////  BK::FontDatabase
//

BK::FontDatabase * BK::FontDatabase::singleton_object = NULL;

BK::FontDatabase::FontDatabase ( void )
        : num_cache_size    (1024),
          num_search_total  (0),
          num_hit_total     (0)
{
    if( singleton_object != NULL ) {
        throw new logic_exception("FontDatabase ̍č\zs܂.");
    }
    singleton_object = this;
    if( log_is_comment() ) {
        log_comment("Singleton FontDatabase <%p> \z.", this);
    }
}

BK::FontDatabase::~FontDatabase ()
{
    if( singleton_object != this ) {
        throw new logic_exception("s FontDatabase o܂.");
    }
    singleton_object = NULL;
    if( log_is_comment() ) {
        log_comment("Singleton FontDatabase <%p> j.", this);
    }
}

void BK::FontDatabase::ruby_gc_mark ( void )
{
    std::for_each( this->map_font.begin(), this->map_font.end(), RubyMapMarker< fontmap_t >() );
    {
        const cachelist_t::iterator end = this->ary_cache.end();
        cachelist_t::iterator i;
        for( i = ary_cache.begin(); i != end; ++i ) {
            Glyph * const glyph = i->get_glyph();
            glyph->gc_mark_self();
        }
    }
}

void BK::FontDatabase::add_font ( const char * name, BK::Font & font )
{
    if( name == NULL ) {
        throw new argument_exception("tHgf[^x[Xւ̒ǉŃtHg NULL w肳܂.");
    }
    if( name[0] == '\0' ) {
        throw new argument_exception("tHgf[^x[Xւ̒ǉŃtHgɋ󕶎񂪎w肳܂.");
    }
    fontmap_t & fontmap = this->map_font;
    const fontmap_t::iterator i = fontmap.find( name );
    
    if( i != fontmap.end() ) {
        if( log_is_info() ) {
            log_info("tHgf[^x[Xɓo^ς݂̃tHg '%s' ㏑܂.", name);
        }
        i->second = &font;
    } else {
        if( !fontmap.insert( fontmap_t::value_type(name, &font) ).second ) {
            throw new runtime_exception("tHgf[^x[Xւ̃tHg '%s' ǉɎs܂.", name);
        }
    }
}

BK::Font * BK::FontDatabase::erase_font ( const char * name )
{
    if( name == NULL ) {
        throw new argument_exception("tHgf[^x[X̍폜 NULL w肳܂.");
    }
    fontmap_t & fontmap = this->map_font;
    const fontmap_t::iterator i = map_font.find( name );
    
    if( i == fontmap.end() ) {
        throw new argument_exception("tHgf[^x[XɃtHg '%s' ͖o^ł.", name);
    }
    Font * const ret = i->second;
    
    fontmap.erase( i );
    
    return ret;
}

BK::Glyph * BK::FontDatabase::get_glyph ( BK::Font & font, const char * str )
{
    if( str == NULL ) {
        throw new argument_exception("tHgf[^x[X̃Ot NULL w肳܂.");
    }
    if( str[0] == '\0' ) {
        return NULL;
    }
    const key_t key ( font, str );
    const cachelist_t::iterator i = std::find( ary_cache.begin(), ary_cache.end(), key );
    
    num_search_total += 1;
    if( i != ary_cache.end() ) {
        // LbVqbg
        num_hit_total += 1;
        const elem_t elem = *i;
        
        ary_cache.erase( i );
        ary_cache.push_front( elem );
        
        return elem.get_glyph();
    }
    else {
        // LbV~X
        Glyph * const glyph = font.create_glyph( str );
        
        if( ary_cache.size() >= num_cache_size ) {
            while( ary_cache.size() > num_cache_size ) {
                ary_cache.pop_back();
            }
            if( num_cache_size > 0 ) {
                ary_cache.pop_back();
            }
        }
        ary_cache.push_back( elem_t(key, glyph) );
        
        return glyph;
    }
}

void BK::FontDatabase::set_cache_size ( unsigned int n )
{
    num_cache_size = n;
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_graphic_Font ( void )
{
    new FontDatabase ();
    
    BK::Font::get_ruby_class();
    BK::Glyph::get_ruby_class();
    BK::GlyphLayout::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_graphic_Font ( void )
{
    if( log_is_info() ) {
        log_info("tHgLbṼqbg ... %.2f%%", FontDatabase::get_instance()->get_cache_hit_percent());
    }
    delete FontDatabase::get_instance();
}

//////////////////////////////////////////////////////////////////////
