#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::SpriteDatabase
//
namespace {
extern "C" {

/// XvCgf[^x[X̐
VALUE SpriteDatabase_new ( VALUE klass )
{
    using namespace BK;
    VALUE volatile obj = Qnil;
    
    try {
        SpriteDatabase * const ptr = new SpriteDatabase ();
        obj = ptr->ruby_wrap();
        rb_obj_call_init( obj, 0, NULL );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return obj;
}

/// IuWFNg̏
VALUE SpriteDatabase_initialize ( VALUE self )
{
    return self;
}

/// XvCgf[^x[XX^bNɃf[^x[Xǉ
VALUE SpriteDatabase_global_push ( VALUE klass, VALUE dbobj )
{
    using namespace BK;
    SpriteDatabaseStack * const stack = SpriteDatabaseStack::get_instance();
    
    if( stack == NULL ) {
        Ruby::raise_object_is_not_created("XvCgf[^x[XX^bN");
    }
    
    if( !SpriteDatabase::is_kind_of(dbobj) ) {
        Ruby::raise_invalid_parameter_type_to_register("XvCgf[^x[XX^bN", "SpriteDatabase", dbobj);
    }
    BK::SpriteDatabase * const db = BK::SpriteDatabase::value_to_data( dbobj );
    
    try {
        stack->push( *db );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return dbobj;
}

/// XvCgf[^x[XX^bN̍ŏʃf[^o
VALUE SpriteDatabase_global_pop ( VALUE klass )
{
    using namespace BK;
    SpriteDatabaseStack * const stack = SpriteDatabaseStack::get_instance();
    
    if( stack == NULL ) {
        Ruby::raise_object_is_not_created("XvCgf[^x[XX^bN");
    }
    
    VALUE volatile obj = Qnil;
    
    try {
        obj = stack->get_top().ruby_wrap();
        stack->pop();
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return obj;
}

/// XvCgf[^x[XX^bNXvCg
VALUE SpriteDatabase_global_find ( VALUE klass, VALUE keyobj )
{
    using namespace BK;
    SpriteDatabaseStack * const stack = SpriteDatabaseStack::get_instance();
    
    if( stack == NULL ) {
        Ruby::raise_object_is_not_created("XvCgf[^x[XX^bN");
    }
    
    const char * const key = StringValuePtr( keyobj );
    Sprite * sprite;
    
    try {
        sprite = stack->find( key );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return (sprite == NULL) ? Qnil : sprite->get_ruby_object();
}

/// XvCgf[^x[XX^bN̍ŏʃf[^x[XɃf[^o^
VALUE SpriteDatabase_global_register ( VALUE klass, VALUE keyobj, VALUE spriteobj )
{
    using namespace BK;
    SpriteDatabaseStack * const stack = SpriteDatabaseStack::get_instance();
    
    if( stack == NULL ) {
        Ruby::raise_object_is_not_created("XvCgf[^x[XX^bN");
    }
    
    if( !Sprite::is_kind_of(spriteobj) ) {
        Ruby::raise_invalid_parameter_type_to_register("XvCgf[^x[X", "Sprite", spriteobj);
    }
    
    SpriteDatabase & db = stack->get_top();
    VALUE volatile dbobj = db.ruby_wrap();
    
    rb_check_frozen( dbobj );
    
    const char * const key = StringValuePtr( keyobj );
    
    try {
        Sprite * const sprite = Sprite::value_to_data( spriteobj );
        
        db.add( key, *sprite );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return spriteobj;
}

/// XvCg̓o^
VALUE SpriteDatabase_register ( VALUE self, VALUE keyobj, VALUE spriteobj )
{
    using namespace BK;
    
    if( !Sprite::is_kind_of(spriteobj) ) {
        Ruby::raise_invalid_parameter_type_to_register("XvCgf[^x[X", "Sprite", spriteobj);
    }
    
    rb_check_frozen( self );
    
    SpriteDatabase * const ptr = SpriteDatabase::value_to_data( self );
    const char * const key = StringValuePtr( keyobj );
    
    try {
        Sprite * const sprite = Sprite::value_to_data( spriteobj );
        
        ptr->add( key, *sprite );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return spriteobj;
}

/// XvCg̎擾
VALUE SpriteDatabase_get_sprite ( VALUE self, VALUE keyobj )
{
    using namespace BK;
    SpriteDatabase * const ptr = SpriteDatabase::value_to_data( self );
    const char * const key = StringValuePtr( keyobj );
    Sprite * sprite;
    
    try {
        sprite = ptr->find( key );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return (sprite == NULL) ? Qnil : sprite->get_ruby_object();
}

/// f[^x[Xɓo^ꂽXvCg̍폜
VALUE SpriteDatabase_delete ( VALUE self, VALUE nameobj )
{
    using namespace BK;
    
    rb_check_frozen( self );
    
    SpriteDatabase * const ptr = SpriteDatabase::value_to_data( self );
    const char * const name = StringValuePtr( nameobj );
    
    try {
        ptr->erase( name );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return Qtrue;
}

/// f[^x[Xɓo^ꂽXvCg̎擾
VALUE SpriteDatabase_get_size ( VALUE self )
{
    using namespace BK;
    SpriteDatabase * const ptr = SpriteDatabase::value_to_data( self );
    
    return rb_uint_new( ptr->size() );
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  SpriteDatabase
//

VALUE BK::SpriteDatabase::ruby_class_object = Qnil;

BK::SpriteDatabase::SpriteDatabase ( void )
{
}

BK::SpriteDatabase::~SpriteDatabase ()
{
}

VALUE BK::SpriteDatabase::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "SpriteDatabase", parent_t::get_ruby_class() );
    
    xrb_define_0arg_singleton_method( c, "new", SpriteDatabase_new );
    xrb_define_1arg_singleton_method( c, "push", SpriteDatabase_global_push );
    xrb_define_0arg_singleton_method( c, "pop", SpriteDatabase_global_pop );
    xrb_define_1arg_singleton_method( c, "[]", SpriteDatabase_global_find );
    xrb_define_2arg_singleton_method( c, "[]=", SpriteDatabase_global_register );
    
    xrb_define_0arg_method( c, "initialize", SpriteDatabase_initialize );
    xrb_define_1arg_method( c, "[]", SpriteDatabase_get_sprite );
    xrb_define_2arg_method( c, "[]=", SpriteDatabase_register );
    xrb_define_1arg_method( c, "delete", SpriteDatabase_delete );
    xrb_define_0arg_method( c, "size", SpriteDatabase_get_size );
    
    return ruby_class_object;
}

void BK::SpriteDatabase::ruby_gc_mark ( void )
{
    parent_t::ruby_gc_mark();
    std::for_each( this->map_sprite.begin(), this->map_sprite.end(), RubyMapMarker< map_t >() );
}

void BK::SpriteDatabase::on_ruby_wrap ( void )
{
    parent_t::on_ruby_wrap();
    std::for_each( this->map_sprite.begin(), this->map_sprite.end(), RubyMapWrapper< map_t >() );
}

void BK::SpriteDatabase::add ( const char * name, BK::Sprite & sprite )
{
    if( name == NULL ) {
        throw new argument_exception(
            "XvCgf[^x[Xւ̓o^ŃXvCg NULL w肳܂.");
    }
    if( name[0] == '\0' ) {
        throw new argument_exception(
            "XvCgf[^x[Xւ̓o^ŃXvCgɋ󕶎񂪎w肳܂.");
    }
    const map_t::iterator i = map_sprite.find( name );
    
    if( i != map_sprite.end() ) {
        log_info("XvCgf[^x[X <%p> ɓo^ς݂̃XvCg '%s' ㏑܂.", this, name);
        map_sprite.erase( i );
    }
    
    if( !map_sprite.insert( map_t::value_type(name, &sprite) ).second ) {
        throw new runtime_exception("XvCgf[^x[XŃXvCg '%s' ̒ǉɎs܂.", name);
    }
    
    if( log_is_trace() ) {
        log_trace("XvCgf[^x[X <%p> ɃXvCg '%s'(img:%s, rect:%s, center:%s) ǉ.",
                  this, name,
                  typeid(*(sprite.get_picture())).name(),
                  sprite.get_rect().to_string().c_str(),
                  sprite.get_center().to_string().c_str());
    }
}

void BK::SpriteDatabase::erase ( const char * name )
{
    if( name == NULL ) {
        throw new argument_exception("XvCgf[^x[X̍폜 NULL w肳܂.");
    }
    const map_t::iterator i = map_sprite.find( name );
    
    if( i == map_sprite.end() ) {
        throw new argument_exception("XvCgf[^x[XŃXvCg '%s' ͖o^ł.", name);
    }
    map_sprite.erase( i );
}

BK::Sprite * BK::SpriteDatabase::find ( const char * name ) const
{
    if( name == NULL ) {
        throw new argument_exception("XvCgf[^x[X̌ŃXvCg NULL w肳܂.");
    }
    const map_t::const_iterator i = map_sprite.find( name );
    
    return (i == map_sprite.end()) ? NULL : i->second;
}

//////////////////////////////////////////////////////////////////////
////  SpriteDatabaseStack
//

BK::SpriteDatabaseStack * BK::SpriteDatabaseStack::singleton_object = NULL;

BK::SpriteDatabaseStack::SpriteDatabaseStack ( void )
{
    if( singleton_object != NULL ) {
        throw new logic_exception("SpriteDatabaseStack ̍č\zs܂.");
    }
    singleton_object = this;
    if( log_is_comment() ) {
        log_comment("Singleton SpriteDatabaseStack <%p> \z.", this);
    }
}

BK::SpriteDatabaseStack::~SpriteDatabaseStack ( void )
{
    if( singleton_object != this ) {
        throw new logic_exception("s SpriteDatabaseStack o܂.");
    }
    singleton_object = NULL;
    if( log_is_comment() ) {
        log_comment("Singleton SpriteDatabaseStack <%p> j.", this);
    }
}

void BK::SpriteDatabaseStack::ruby_gc_mark ( void )
{
    std::for_each( this->ary_database.begin(), this->ary_database.end(), RubyObjectMarker() );
    this->obj_default_database.ruby_gc_mark();
}

BK::Sprite * BK::SpriteDatabaseStack::find ( const char * name ) const
{
    const list_t::const_reverse_iterator end = this->ary_database.rend();
    list_t::const_reverse_iterator i;
    
    for( i = this->ary_database.rbegin(); i != end; ++i ) {
        Sprite * const ptr = (*i)->find( name );
        
        if( ptr != NULL ) {
            return ptr;
        }
    }
    return obj_default_database.find( name );
}

void BK::SpriteDatabaseStack::push ( BK::SpriteDatabase & db )
{
    // O蓖Ă߂prl̂œ DB ̒ǉ͕sȏł͂Ȃ
    //if( std::find(ary_database.begin(), ary_database.end(), db) != ary_database.end() ) {
    //    log_warn("XvCg DB X^bNɓ DB (%s) ǉ܂.", db->get_name());
    //}
    ary_database.push_back( &db );
}

BK::SpriteDatabase * BK::SpriteDatabaseStack::pop ( void )
{
    if( ary_database.empty() ) {
        throw new runtime_exception("̃XvCgf[^x[XX^bNō폜s܂.");
    }
    SpriteDatabase * const ret = ary_database.back();
    
    ary_database.pop_back();
    
    return ret;
}

BK::SpriteDatabase & BK::SpriteDatabaseStack::get_top ( void )
{
    if( ary_database.empty() ) {
        return obj_default_database;
    } else {
        return *(ary_database.back());
    }
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_graphic_SpriteDatabase ( void )
{
    new SpriteDatabaseStack ();
    SpriteDatabase::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_graphic_SpriteDatabase ( void )
{
    delete SpriteDatabaseStack::get_instance();
}

//////////////////////////////////////////////////////////////////////
