#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::SpriteSet
//
namespace {
extern "C" {

/// XvCgZbg̐
VALUE SpriteSet_new ( int argc, VALUE * argv, VALUE klass )
{
    using namespace BK;
    VALUE volatile obj = Qnil;
    
    try {
        Ruby::ArgScanner scanner("SpriteSet.new", argc, argv);
        Size2 size;
        
        size = scanner.fetch_Size2("̈TCY");
        
        scanner.finish();
        
        SpriteSet * const ptr = new SpriteSet ( size );
        
        obj = ptr->ruby_wrap();
        rb_obj_call_init( obj, argc, argv );
        
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return obj;
}

/// IuWFNg̏
VALUE SpriteSet_initialize ( int argc, VALUE * argv, VALUE self )
{
    return self;
}

/// O[oXvCgZbgɃXvCgZbgo^
VALUE SpriteSet_global_add ( VALUE klass, VALUE nameobj, VALUE setobj )
{
    using namespace BK;
    GlobalSpriteSet * const ptr = GlobalSpriteSet::get_instance();
    
    if( ptr == NULL ) {
        Ruby::raise_object_is_not_created("O[oXvCgZbg");
    }
    
    if( !SpriteSet::is_kind_of(setobj) ) {
        Ruby::raise_invalid_parameter_type_to_register("O[oXvCgZbg", "SpriteSet", setobj);
    }
    SpriteSet * const set = SpriteSet::value_to_data( setobj );
    const char * const name = StringValuePtr( nameobj );
    
    try {
        ptr->add( name, *set );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return Qtrue;
}

/// O[oXvCgZbgXvCgZbg擾
VALUE SpriteSet_global_find ( VALUE klass, VALUE nameobj )
{
    using namespace BK;
    GlobalSpriteSet * const ptr = GlobalSpriteSet::get_instance();
    
    if( ptr == NULL ) {
        Ruby::raise_object_is_not_created("O[oXvCgZbg");
    }
    const char * const name = StringValuePtr( nameobj );
    SpriteSet * set;
    
    try {
        set = ptr->find( name );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return (set == NULL) ? Qnil : set->get_ruby_object();
}

/// ZbgɃf[^ǉ
VALUE SpriteSet_add ( int argc, VALUE * argv, VALUE self )
{
    using namespace BK;
    SpriteSet * const ptr = SpriteSet::value_to_data( self );
    VALUE nameobj;
    VALUE areaobj;
    Point2 point;
    
    rb_check_frozen( self );
    
    try {
        Ruby::ArgScanner scanner("Sprite#add", argc, argv);
        
        nameobj = scanner.fetch_value("XvCgZbg");
        areaobj = scanner.fetch_value("̈IuWFNg");
        point   = scanner.fetch_Point2("̈ʒu");
        
        scanner.finish();
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    Sprite * sprite = NULL;
    SpriteSet * set = NULL;
    
    if( Sprite::is_kind_of(areaobj) ) {
        sprite = Sprite::value_to_data(areaobj);
    }
    else if( SpriteSet::is_kind_of(areaobj) ) {
        set = SpriteSet::value_to_data(areaobj);
    }
    else {
        rb_raise(rb_eArgError, "XvCgZbg %s ͓o^s.", rb_obj_classname(areaobj));
    }
    
    const char * name = StringValuePtr( nameobj );
    
    try {
        if( sprite != NULL ) {
            ptr->add_sprite( name, *sprite, point );
        } else {
            ptr->add_set( name, *set, point );
        }
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return Qtrue;
}

/// o^ꂽf[^폜
VALUE SpriteSet_delete ( VALUE self, VALUE nameobj )
{
    using namespace BK;
    
    rb_check_frozen( self );
    
    SpriteSet * const ptr = SpriteSet::value_to_data( self );
    const char * name = StringValuePtr( nameobj );
    VALUE obj;
    
    try {
        obj = ptr->erase( name ).ruby_wrap();
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return obj;
}

/// o^ꂽf[^擾
VALUE SpriteSet_find ( VALUE self, VALUE nameobj )
{
    using namespace BK;
    SpriteSet * const ptr = SpriteSet::value_to_data( self );
    SpriteSet::data_t const * data;
    const char * name = StringValuePtr( nameobj );
    
    try {
        data = ptr->find( name );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return (data->is_sprite()) ? data->get_sprite()->get_ruby_object() : data->get_set()->get_ruby_object();
}

/// XvCgZbg̗̈TCY擾
VALUE SpriteSet_get_area ( VALUE self )
{
    BK::SpriteSet * const ptr = BK::SpriteSet::value_to_data( self );
    
    return ptr->get_size().create_ruby_object();
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  SpriteSet
//

VALUE BK::SpriteSet::ruby_class_object = Qnil;

BK::SpriteSet::SpriteSet ( BK::Size2 const & size )
        : size_field  (size),
          num_sprite  (0)
{
}

BK::SpriteSet::~SpriteSet ()
{
    for( map_t::iterator i = map_data.begin(); i != map_data.end(); ++i ) {
        delete i->second;
    }
}

VALUE BK::SpriteSet::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "SpriteSet", parent_t::get_ruby_class() );
    
    xrb_define_narg_singleton_method( c, "new", SpriteSet_new );
    xrb_define_1arg_singleton_method( c, "[]", SpriteSet_global_find );
    xrb_define_2arg_singleton_method( c, "[]=", SpriteSet_global_add );
    
    xrb_define_narg_method( c, "initialize", SpriteSet_initialize );
    xrb_define_narg_method( c, "add", SpriteSet_add );
    xrb_define_1arg_method( c, "delete", SpriteSet_delete );
    xrb_define_1arg_method( c, "[]", SpriteSet_find );
    xrb_define_0arg_method( c, "area", SpriteSet_get_area );
    
    return ruby_class_object;
}

void BK::SpriteSet::ruby_gc_mark ( void )
{
    parent_t::ruby_gc_mark();
    
    const map_t::iterator e = this->map_data.end();
    map_t::iterator itr;
    
    for( itr = this->map_data.begin(); itr != e; ++itr ) {
        data_t * const dat = itr->second;
        dat->gc_mark_self(); // data_t  RubyObject łȂ̂ RubyMapMarker<> gȂ
    }
}

void BK::SpriteSet::on_ruby_wrap ( void )
{
    parent_t::on_ruby_wrap();
    
    const map_t::iterator e = this->map_data.end();
    map_t::iterator itr;
    
    for( itr = this->map_data.begin(); itr != e; ++itr ) {
        data_t * const dat = itr->second;
        dat->on_ruby_wrap(); // data_t  RubyObject łȂ̂ RubyMapWrapper<> gȂ
    }
}

bool BK::SpriteSet::is_valid_rect ( const Rect2 & r ) const
{
    // Zbg̈Ɏ܂ĂȂ΂ȂȂ
    if( !Rect2(this->get_size()).is_include( r ) ) {
        return false;
    }
    
    // ǂ̃GAƂdȂĂ͂ȂȂ
    const map_t::const_iterator e = this->map_data.end();
    map_t::const_iterator itr;
    
    for( itr = this->map_data.begin(); itr != e; ++itr ) {
        data_t * const dat = itr->second;
        
        if( dat->get_area().is_shared( r ) ) {
            return false;
        }
    }
    
    return true;
}

void BK::SpriteSet::add_data ( const char * name, const char * datatype, BK::SpriteSet::data_t * data )
{
    const Rect2 area = data->get_area();
    
    if( !is_valid_rect(area) ) {
        throw new argument_exception(
            "XvCgZbg̈ʒu %s %s \"%s\" ǉł܂.",
            area.get_point().to_string().c_str(), datatype, name);
    }
    const map_t::iterator itr = this->map_data.find( name );
    
    if( itr != this->map_data.end() ) {
        log_info("XvCgZbgɓo^ς݂̃f[^ \"%s\" ㏑܂.", name);
        delete itr->second;
        itr->second = data;
    }
    else {
        if( !this->map_data.insert( map_t::value_type(name, data) ).second ) {
            throw new runtime_exception("XvCgZbg%s \"%s\" ̒ǉɎs܂.", datatype, name);
        }
    }
}

void BK::SpriteSet::add_sprite ( const char * name, BK::Sprite & sprite, BK::Point2 const & p )
{
    if( name == NULL ) {
        throw new argument_exception("XvCgZbgւ̃XvCgo^Ńf[^ NULL w肳܂.");
    }
    if( name[0] == '\0' ) {
        throw new argument_exception("XvCgZbgւ̃XvCgo^ŋ̃f[^w肳܂.");
    }
    this->add_data( name, "XvCg", new data_t(sprite, p) );
    this->num_sprite += 1;
}

void BK::SpriteSet::add_set ( const char * name, BK::SpriteSet & set, BK::Point2 const & p )
{
    if( name == NULL ) {
        throw new argument_exception("XvCgZbgւ̃XvCgZbgo^Ńf[^ NULL w肳܂.");
    }
    if( name[0] == '\0' ) {
        throw new argument_exception("XvCgZbgւ̃XvCgZbgo^ŋ̃f[^w肳܂.");
    }
    if( &set == this ) {
        throw new argument_exception("XvCgZbgւ̃XvCgZbgo^Ŏgw肳܂.");
    }
    this->add_data( name, "XvCgZbg", new data_t(set, p) );
}

BK::SpriteSet::data_t BK::SpriteSet::erase ( const char * name )
{
    if( name == NULL ) {
        throw new argument_exception("XvCgZbg̍폜Ńf[^ NULL w肳܂.");
    }
    const map_t::iterator itr = this->map_data.find( name );
    
    if( itr == map_data.end() ) {
        throw new argument_exception("\"%s\" ̓XvCgZbgɑ݂܂.", name);
    }
    data_t * const erased = itr->second;
    const data_t ret = *erased;
    
    delete erased;
    map_data.erase( itr );
    if( ret.is_sprite() ) {
        this->num_sprite -= 1;
    }
    
    return ret;
}

BK::SpriteSet::data_t const * BK::SpriteSet::find ( const char * name ) const
{
    if( name == NULL ) {
        throw new argument_exception("XvCgZbǧ NULL w肳܂.");
    }
    if( strlen(name) >= MAX_SPRITEPATH_STRING ) {
        throw new argument_exception("XvCgZbǧŎw肳ꂽpX񂪒܂.");
    }
    char buf [ MAX_SPRITEPATH_STRING ];
    
    strcpy( buf, name );
    
    const unsigned int count = str_translate_char(buf, '/', '\0') + 1;
    
    if( count > MAX_SPRITEPATH_NEST ) {
        throw new argument_exception("XvCgZbǧŎw肳ꂽpX̊Kw[܂.");
    }
    
    char * ary [ MAX_SPRITEPATH_NEST ];
    {
        char * * ptr = ary;
        char * p = buf;
        
        for( unsigned int i = 1; i < count; ++i ) {
            *(ptr++) = p;
            p += strlen(p) + 1;
        }
        *ptr = p;
    }
    
    return this->find(count, ary);
}

BK::SpriteSet::data_t const * BK::SpriteSet::find ( unsigned int length, const char * const * names ) const
{
    SpriteSet const * currentset = this;
    data_t const * data = NULL;
    
    while( length > 0 ) {
        if( currentset == NULL ) {
            return NULL;
        }
        data = currentset->get_data( *names );
        if( data == NULL ) {
            return NULL;
        }
        currentset = data->get_set();
        --length;
        ++names;
    }
    return data;
}

BK::SpriteSet::data_t const * BK::SpriteSet::get_data ( const char * name ) const
{
    const map_t::const_iterator i = map_data.find( name );
    
    return (i == map_data.end()) ? NULL : i->second;
}

unsigned int BK::SpriteSet::count_sprite ( void ) const
{
    const map_t::const_iterator end = this->map_data.end();
    map_t::const_iterator i;
    unsigned int count = this->num_sprite;
    
    for( i = this->map_data.begin(); i != end; ++i ) {
        const data_t & data = *(i->second);
        
        if( data.is_set() ) {
            count += data.get_set()->count_sprite();
        }
    }
    
    return count;
}

//////////////////////////////////////////////////////////////////////
////  GlobalSpriteSet
//

BK::GlobalSpriteSet * BK::GlobalSpriteSet::singleton_object = NULL;

BK::GlobalSpriteSet::GlobalSpriteSet ( void )
{
    if( singleton_object != NULL ) {
        throw new logic_exception("GlobalSpriteSet ̍č\zs܂.");
    }
    singleton_object = this;
}

BK::GlobalSpriteSet::~GlobalSpriteSet ()
{
    if( singleton_object != this ) {
        throw new logic_exception("s GlobalSpriteSet o܂.");
    }
    singleton_object = NULL;
}

void BK::GlobalSpriteSet::ruby_gc_mark ( void )
{
    std::for_each( map_spriteset.begin(), map_spriteset.end(), RubyMapMarker< map_t >() );
}

BK::SpriteSet * BK::GlobalSpriteSet::find ( const char * name )
{
    if( name == NULL ) {
        throw new argument_exception("O[oXvCgZbǧ NULL w肳܂.");
    }
    if( name[0] != '/' ) {
        throw new argument_exception(
            "O[oXvCgZbǧɑ΃pX (%s) w肳܂.", name);
    }
    ++name;
    
    char * buf = ALLOCA_N(char, strlen(name) + 1);
    strcpy( buf, name );
    
    unsigned int const count = str_translate_char(buf, '/', '\0') + 1;
    
    char * * const ary = ALLOCA_N(char *, count);
    
    char * * ptr = ary;
    
    for( unsigned int i = 1; i < count; ++i ) {
        *(ptr++) = buf;
        buf += strlen(buf) + 1;
    }
    *ptr = buf;
    
    return find(count, ary);
}

BK::SpriteSet * BK::GlobalSpriteSet::find ( unsigned int length, const char * const * names )
{
    if( length == 0 ) {
        return NULL;
    }
    
    map_t::iterator const i = map_spriteset.find( *names );
    if( i == map_spriteset.end() ) {
        return NULL;
    }
    SpriteSet * currentset = i->second;
    
    ++names;
    --length;
    
    SpriteSet::data_t const * data = NULL;
    
    while( length > 0 ) {
        if( currentset == NULL ) {
            return NULL;
        }
        data = currentset->get_data( *names );
        if( data == NULL ) {
            return NULL;
        }
        currentset = data->get_set();
        --length;
    }
    return (data == NULL) ? NULL : data->get_set();
}

void BK::GlobalSpriteSet::add ( const char * name, BK::SpriteSet & set )
{
    if( name == NULL ) {
        throw new argument_exception("O[oXvCgZbgւ̒ǉŃXvCgZbg NULL w肳܂.");
    }
    if( name[0] == '\0' ) {
        throw new argument_exception("O[oXvCgZbgւ̒ǉŃXvCgZbgɋ󕶎񂪎w肳܂.");
    }
    const map_t::iterator itr = map_spriteset.find( name );
    
    if( itr != map_spriteset.end() ) {
        log_info("O[oXvCgZbgɓo^ς݂̃XvCgZbg '%s' ㏑܂.", name);
        map_spriteset.erase( itr );
    }
    if( !map_spriteset.insert( map_t::value_type(name, &set) ).second ) {
        throw new runtime_exception("O[oXvCgZbgւ̃XvCgZbg '%s' ǉɎs܂.", name);
    }
}

BK::SpriteSet * BK::GlobalSpriteSet::erase ( const char * name )
{
    if( name == NULL ) {
        throw new argument_exception("O[oXvCgZbg̍폜 NULL w肳܂.");
    }
    const map_t::iterator i = map_spriteset.find( name );
    
    if( i == map_spriteset.end() ) {
        throw new argument_exception("O[oXvCgZbgɃXvCgZbg '%s' ͖o^ł.", name);
    }
    SpriteSet * const ret = i->second;
    
    map_spriteset.erase( i );
    
    return ret;
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_graphic_SpriteSet ( void )
{
    new GlobalSpriteSet ();
    SpriteSet::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_graphic_SpriteSet ( void )
{
    delete GlobalSpriteSet::get_instance();
}

//////////////////////////////////////////////////////////////////////
