#include "stdafx.h"
#include "schema/main.h"

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::TextLabel
//
namespace {
extern "C" {

/// Kx[WRN^p}[N֐
void TextLabel_mark ( BK::TextLabel * ptr )
{
    ptr->ruby_gc_mark();
}

/// IuWFNg̉
void TextLabel_free ( BK::TextLabel * ptr )
{
    delete ptr;
}

/// x̎擾
VALUE TextLabel_get_label ( VALUE self )
{
    BK::TextLabel * const ptr = BK::TextLabel::value_to_data( self );
    
    return rb_str_new2(ptr->get_label());
}

/// x\`̎擾
VALUE TextLabel_get_area ( VALUE self )
{
    BK::TextLabel * const ptr = BK::TextLabel::value_to_data( self );
    
    return ptr->get_area().create_ruby_object();
}

/// x̐ݒ
VALUE TextLabel_set_label ( VALUE self, VALUE strobj )
{
    BK::TextLabel * const ptr = BK::TextLabel::value_to_data( self );
    const char * str = StringValuePtr( strobj );
    
    return ptr->set_label( str );
}

/// x\`̐ݒ
VALUE TextLabel_set_area ( VALUE self, VALUE areaobj )
{
    using namespace BK;
    TextLabel * const ptr = TextLabel::value_to_data( self );
    
    try {
        ptr->set_area( value_to_Rect2(areaobj) );
    } catch ( exception_base * e ) {
        ruby_throw( e );
    }
    
    return areaobj;
}

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::TextLabel
//

VALUE BK::TextLabel::ruby_class_object = Qnil;

VALUE BK::TextLabel::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "TextLabel", rb_cData );
    
    xrb_define_0arg_method( c, "label", TextLabel_get_label );
    xrb_define_0arg_method( c, "area", TextLabel_get_area );
    xrb_define_1arg_method( c, "label=", TextLabel_set_label );
    xrb_define_1arg_method( c, "area=", TextLabel_set_area );
    
    return ruby_class_object;
}

BK::TextLabel::TextLabel ( void )
        : ptr_font (NULL)
{
}

BK::TextLabel::~TextLabel ()
{
}

void * BK::TextLabel::get_ruby_mark_function ( void )
{
    return TextLabel_mark;
}

void * BK::TextLabel::get_ruby_free_function ( void )
{
    return TextLabel_free;
}

void BK::TextLabel::ruby_gc_mark ( void )
{
    RubyObject::ruby_gc_mark();
    if( ptr_font != NULL ) {
        ptr_font->ruby_gc_mark();
    }
}

void BK::TextLabel::set_label ( const char * str )
{
    if( str == NULL ) {
        if( str_label.length() > 0 ) {
            str_label = str;
            update_layout();
        }
    } else {
        if( str_label != str ) {
            str_label = str;
            update_layout();
        }
    }
}

void BK::TextLabel::set_area ( const BK::Rect2 & area )
{
    const Size2 prevsize = rect_area.get_size();
    
    rect_area = area;
    if( prevsize != area.get_size() ) {
        update_layout();
    }
}

void BK::TextLabel::set_font ( BK::Font * font )
{
    if( ptr_font != font ) {
        ptr_font = font;
        update_layout();
    }
}

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::SimpleTextLabel
//
namespace {
extern "C" {

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::SimpleTextLabel
//

VALUE BK::SimpleTextLabel::ruby_class_object = Qnil;

VALUE BK::SimpleTextLabel::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "SimpleTextLabel", BK::TextLabel::get_ruby_class() );
    
    xrb_define_0arg_method( c, "horizontal", SimpleTextLabel_get_horizontal );
    xrb_define_0arg_method( c, "vertical", SimpleTextLabel_get_vertical );
    xrb_define_1arg_method( c, "horizontal=", SimpleTextLabel_set_horizontal );
    xrb_define_1arg_method( c, "vertical=", SimpleTextLabel_set_vertical );
    
    return ruby_class_object;
}

BK::SimpleTextLabel::SimpleTextLabel ( void )
        : num_horizontal (LEFT),
          num_vertical (TOP)
{
}

BK::SimpleTextLabel::~SimpleTextLabel ()
{
}

void BK::SimpleTextLabel::ruby_gc_mark ( void )
{
    const list_t::iterator end = ary_info.end();
    list_t::iterator i;
    
    for( i = ary_info.begin(); i != end; ++i ) {
        i->ruby_gc_mark();
    }
}

void BK::SimpleTextLabel::set_horizontal ( BK::SimpleTextLabel::align_t align )
{
}

void BK::SimpleTextLabel::set_vertical ( BK::SimpleTextLabel::align_t align )
{
}

void BK::SimpleTextLabel::update_layout ( void )
{
}

void BK::SimpleTextLabel::execute ( BK::GraphicDevice * dev )
{
}

//////////////////////////////////////////////////////////////////////
////  Ruby IuWFNg: BK::TextLabelLayer
//
namespace {
extern "C" {

} /* extern "C" */
} /* namespace */
//////////////////////////////////////////////////////////////////////
////  BK::TextLabelLayer
//

VALUE BK::TextLabelLayer::ruby_class_object = Qnil;

VALUE BK::TextLabelLayer::get_ruby_class ( void )
{
    if( !NIL_P(ruby_class_object) ) {
        return ruby_class_object;
    }
    VALUE c;
    
    c = ruby_class_object = define_ruby_bk_class( "TextLabelLayer", parent_t::get_ruby_class() );
    
    return ruby_class_object;
}

BK::TextLabelLayer::TextLabelLayer ( void )
{
}

BK::TextLabelLayer::~TextLabelLayer ()
{
}

void BK::TextLabelLayer::ruby_gc_mark ( void )
{
    parent_t::ruby_gc_mark();
}

BK::TextLabel * BK::TextLabelLayer::create_label ( void )
{
    return new TextLabel ();
}

BK::TextLabel * BK::TextLabelLayer::add ( void )
{
    TextLabel * label = create_label();
    
    ary_label.push_back( label );
    
    return label;
}

void BK::TextLabel::erase ( BK::TextLabel * label )
{
    const list_t::iterator i = std::find( ary_label, label );
    
    if( i != ary_label.end() ) {
        throw new argument_exception("w肳ꂽeLXgx̓C[ɖo^ł.");
    }
    ary_label.erase( i );
}

//////////////////////////////////////////////////////////////////////
////  Init ֐
//

void BK::Init_graphic_layer_TextLabelLayer ( void )
{
    BK::TextLabel::get_ruby_class();
    BK::TextLabelLayer::get_ruby_class();
}

//////////////////////////////////////////////////////////////////////
////  Exit ֐
//

void BK::Exit_graphic_layer_TextLabelLayer ( void )
{
}

//////////////////////////////////////////////////////////////////////
